# configitems.py - centralized declaration of configuration option
#
#  Copyright 2017 Pierre-Yves David <pierre-yves.david@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import absolute_import

import functools
import re

from . import (
    encoding,
    error,
)


def loadconfigtable(ui, extname, configtable):
    """update config item known to the ui with the extension ones"""
    for section, items in sorted(configtable.items()):
        knownitems = ui._knownconfig.setdefault(section, itemregister())
        knownkeys = set(knownitems)
        newkeys = set(items)
        for key in sorted(knownkeys & newkeys):
            msg = b"extension '%s' overwrite config item '%s.%s'"
            msg %= (extname, section, key)
            ui.develwarn(msg, config=b'warn-config')

        knownitems.update(items)


class configitem(object):
    """represent a known config item

    :section: the official config section where to find this item,
       :name: the official name within the section,
    :default: default value for this item,
    :alias: optional list of tuples as alternatives,
    :generic: this is a generic definition, match name using regular expression.
    """

    def __init__(
        self,
        section,
        name,
        default=None,
        alias=(),
        generic=False,
        priority=0,
        experimental=False,
    ):
        self.section = section
        self.name = name
        self.default = default
        self.alias = list(alias)
        self.generic = generic
        self.priority = priority
        self.experimental = experimental
        self._re = None
        if generic:
            self._re = re.compile(self.name)


class itemregister(dict):
    """A specialized dictionary that can handle wild-card selection"""

    def __init__(self):
        super(itemregister, self).__init__()
        self._generics = set()

    def update(self, other):
        super(itemregister, self).update(other)
        self._generics.update(other._generics)

    def __setitem__(self, key, item):
        super(itemregister, self).__setitem__(key, item)
        if item.generic:
            self._generics.add(item)

    def get(self, key):
        baseitem = super(itemregister, self).get(key)
        if baseitem is not None and not baseitem.generic:
            return baseitem

        # search for a matching generic item
        generics = sorted(self._generics, key=(lambda x: (x.priority, x.name)))
        for item in generics:
            # we use 'match' instead of 'search' to make the matching simpler
            # for people unfamiliar with regular expression. Having the match
            # rooted to the start of the string will produce less surprising
            # result for user writing simple regex for sub-attribute.
            #
            # For example using "color\..*" match produces an unsurprising
            # result, while using search could suddenly match apparently
            # unrelated configuration that happens to contains "color."
            # anywhere. This is a tradeoff where we favor requiring ".*" on
            # some match to avoid the need to prefix most pattern with "^".
            # The "^" seems more error prone.
            if item._re.match(key):
                return item

        return None


coreitems = {}


def _register(configtable, *args, **kwargs):
    item = configitem(*args, **kwargs)
    section = configtable.setdefault(item.section, itemregister())
    if item.name in section:
        msg = b"duplicated config item registration for '%s.%s'"
        raise error.ProgrammingError(msg % (item.section, item.name))
    section[item.name] = item


# special value for case where the default is derived from other values
dynamicdefault = object()

# Registering actual config items


def getitemregister(configtable):
    f = functools.partial(_register, configtable)
    # export pseudo enum as configitem.*
    f.dynamicdefault = dynamicdefault
    return f


coreconfigitem = getitemregister(coreitems)


def _registerdiffopts(section, configprefix=b''):
    coreconfigitem(
        section,
        configprefix + b'nodates',
        default=False,
    )
    coreconfigitem(
        section,
        configprefix + b'showfunc',
        default=False,
    )
    coreconfigitem(
        section,
        configprefix + b'unified',
        default=None,
    )
    coreconfigitem(
        section,
        configprefix + b'git',
        default=False,
    )
    coreconfigitem(
        section,
        configprefix + b'ignorews',
        default=False,
    )
    coreconfigitem(
        section,
        configprefix + b'ignorewsamount',
        default=False,
    )
    coreconfigitem(
        section,
        configprefix + b'ignoreblanklines',
        default=False,
    )
    coreconfigitem(
        section,
        configprefix + b'ignorewseol',
        default=False,
    )
    coreconfigitem(
        section,
        configprefix + b'nobinary',
        default=False,
    )
    coreconfigitem(
        section,
        configprefix + b'noprefix',
        default=False,
    )
    coreconfigitem(
        section,
        configprefix + b'word-diff',
        default=False,
    )


coreconfigitem(
    b'alias',
    b'.*',
    default=dynamicdefault,
    generic=True,
)
coreconfigitem(
    b'auth',
    b'cookiefile',
    default=None,
)
_registerdiffopts(section=b'annotate')
# bookmarks.pushing: internal hack for discovery
coreconfigitem(
    b'bookmarks',
    b'pushing',
    default=list,
)
# bundle.mainreporoot: internal hack for bundlerepo
coreconfigitem(
    b'bundle',
    b'mainreporoot',
    default=b'',
)
coreconfigitem(
    b'censor',
    b'policy',
    default=b'abort',
    experimental=True,
)
coreconfigitem(
    b'chgserver',
    b'idletimeout',
    default=3600,
)
coreconfigitem(
    b'chgserver',
    b'skiphash',
    default=False,
)
coreconfigitem(
    b'cmdserver',
    b'log',
    default=None,
)
coreconfigitem(
    b'cmdserver',
    b'max-log-files',
    default=7,
)
coreconfigitem(
    b'cmdserver',
    b'max-log-size',
    default=b'1 MB',
)
coreconfigitem(
    b'cmdserver',
    b'max-repo-cache',
    default=0,
    experimental=True,
)
coreconfigitem(
    b'cmdserver',
    b'message-encodings',
    default=list,
)
coreconfigitem(
    b'cmdserver',
    b'track-log',
    default=lambda: [b'chgserver', b'cmdserver', b'repocache'],
)
coreconfigitem(
    b'cmdserver',
    b'shutdown-on-interrupt',
    default=True,
)
coreconfigitem(
    b'color',
    b'.*',
    default=None,
    generic=True,
)
coreconfigitem(
    b'color',
    b'mode',
    default=b'auto',
)
coreconfigitem(
    b'color',
    b'pagermode',
    default=dynamicdefault,
)
coreconfigitem(
    b'command-templates',
    b'graphnode',
    default=None,
    alias=[(b'ui', b'graphnodetemplate')],
)
coreconfigitem(
    b'command-templates',
    b'log',
    default=None,
    alias=[(b'ui', b'logtemplate')],
)
coreconfigitem(
    b'command-templates',
    b'mergemarker',
    default=(
        b'{node|short} '
        b'{ifeq(tags, "tip", "", '
        b'ifeq(tags, "", "", "{tags} "))}'
        b'{if(bookmarks, "{bookmarks} ")}'
        b'{ifeq(branch, "default", "", "{branch} ")}'
        b'- {author|user}: {desc|firstline}'
    ),
    alias=[(b'ui', b'mergemarkertemplate')],
)
coreconfigitem(
    b'command-templates',
    b'pre-merge-tool-output',
    default=None,
    alias=[(b'ui', b'pre-merge-tool-output-template')],
)
coreconfigitem(
    b'command-templates',
    b'oneline-summary',
    default=None,
)
coreconfigitem(
    b'command-templates',
    b'oneline-summary.*',
    default=dynamicdefault,
    generic=True,
)
_registerdiffopts(section=b'commands', configprefix=b'commit.interactive.')
coreconfigitem(
    b'commands',
    b'commit.post-status',
    default=False,
)
coreconfigitem(
    b'commands',
    b'grep.all-files',
    default=False,
    experimental=True,
)
coreconfigitem(
    b'commands',
    b'merge.require-rev',
    default=False,
)
coreconfigitem(
    b'commands',
    b'push.require-revs',
    default=False,
)
coreconfigitem(
    b'commands',
    b'resolve.confirm',
    default=False,
)
coreconfigitem(
    b'commands',
    b'resolve.explicit-re-merge',
    default=False,
)
coreconfigitem(
    b'commands',
    b'resolve.mark-check',
    default=b'none',
)
_registerdiffopts(section=b'commands', configprefix=b'revert.interactive.')
coreconfigitem(
    b'commands',
    b'show.aliasprefix',
    default=list,
)
coreconfigitem(
    b'commands',
    b'status.relative',
    default=False,
)
coreconfigitem(
    b'commands',
    b'status.skipstates',
    default=[],
    experimental=True,
)
coreconfigitem(
    b'commands',
    b'status.terse',
    default=b'',
)
coreconfigitem(
    b'commands',
    b'status.verbose',
    default=False,
)
coreconfigitem(
    b'commands',
    b'update.check',
    default=None,
)
coreconfigitem(
    b'commands',
    b'update.requiredest',
    default=False,
)
coreconfigitem(
    b'committemplate',
    b'.*',
    default=None,
    generic=True,
)
coreconfigitem(
    b'convert',
    b'bzr.saverev',
    default=True,
)
coreconfigitem(
    b'convert',
    b'cvsps.cache',
    default=True,
)
coreconfigitem(
    b'convert',
    b'cvsps.fuzz',
    default=60,
)
coreconfigitem(
    b'convert',
    b'cvsps.logencoding',
    default=None,
)
coreconfigitem(
    b'convert',
    b'cvsps.mergefrom',
    default=None,
)
coreconfigitem(
    b'convert',
    b'cvsps.mergeto',
    default=None,
)
coreconfigitem(
    b'convert',
    b'git.committeractions',
    default=lambda: [b'messagedifferent'],
)
coreconfigitem(
    b'convert',
    b'git.extrakeys',
    default=list,
)
coreconfigitem(
    b'convert',
    b'git.findcopiesharder',
    default=False,
)
coreconfigitem(
    b'convert',
    b'git.remoteprefix',
    default=b'remote',
)
coreconfigitem(
    b'convert',
    b'git.renamelimit',
    default=400,
)
coreconfigitem(
    b'convert',
    b'git.saverev',
    default=True,
)
coreconfigitem(
    b'convert',
    b'git.similarity',
    default=50,
)
coreconfigitem(
    b'convert',
    b'git.skipsubmodules',
    default=False,
)
coreconfigitem(
    b'convert',
    b'hg.clonebranches',
    default=False,
)
coreconfigitem(
    b'convert',
    b'hg.ignoreerrors',
    default=False,
)
coreconfigitem(
    b'convert',
    b'hg.preserve-hash',
    default=False,
)
coreconfigitem(
    b'convert',
    b'hg.revs',
    default=None,
)
coreconfigitem(
    b'convert',
    b'hg.saverev',
    default=False,
)
coreconfigitem(
    b'convert',
    b'hg.sourcename',
    default=None,
)
coreconfigitem(
    b'convert',
    b'hg.startrev',
    default=None,
)
coreconfigitem(
    b'convert',
    b'hg.tagsbranch',
    default=b'default',
)
coreconfigitem(
    b'convert',
    b'hg.usebranchnames',
    default=True,
)
coreconfigitem(
    b'convert',
    b'ignoreancestorcheck',
    default=False,
    experimental=True,
)
coreconfigitem(
    b'convert',
    b'localtimezone',
    default=False,
)
coreconfigitem(
    b'convert',
    b'p4.encoding',
    default=dynamicdefault,
)
coreconfigitem(
    b'convert',
    b'p4.startrev',
    default=0,
)
coreconfigitem(
    b'convert',
    b'skiptags',
    default=False,
)
coreconfigitem(
    b'convert',
    b'svn.debugsvnlog',
    default=True,
)
coreconfigitem(
    b'convert',
    b'svn.trunk',
    default=None,
)
coreconfigitem(
    b'convert',
    b'svn.tags',
    default=None,
)
coreconfigitem(
    b'convert',
    b'svn.branches',
    default=None,
)
coreconfigitem(
    b'convert',
    b'svn.startrev',
    default=0,
)
coreconfigitem(
    b'convert',
    b'svn.dangerous-set-commit-dates',
    default=False,
)
coreconfigitem(
    b'debug',
    b'dirstate.delaywrite',
    default=0,
)
coreconfigitem(
    b'debug',
    b'revlog.verifyposition.changelog',
    default=b'',
)
coreconfigitem(
    b'defaults',
    b'.*',
    default=None,
    generic=True,
)
coreconfigitem(
    b'devel',
    b'all-warnings',
    default=False,
)
coreconfigitem(
    b'devel',
    b'bundle2.debug',
    default=False,
)
coreconfigitem(
    b'devel',
    b'bundle.delta',
    default=b'',
)
coreconfigitem(
    b'devel',
    b'cache-vfs',
    default=None,
)
coreconfigitem(
    b'devel',
    b'check-locks',
    default=False,
)
coreconfigitem(
    b'devel',
    b'check-relroot',
    default=False,
)
# Track copy information for all file, not just "added" one (very slow)
coreconfigitem(
    b'devel',
    b'copy-tracing.trace-all-files',
    default=False,
)
coreconfigitem(
    b'devel',
    b'default-date',
    default=None,
)
coreconfigitem(
    b'devel',
    b'deprec-warn',
    default=False,
)
coreconfigitem(
    b'devel',
    b'disableloaddefaultcerts',
    default=False,
)
coreconfigitem(
    b'devel',
    b'warn-empty-changegroup',
    default=False,
)
coreconfigitem(
    b'devel',
    b'legacy.exchange',
    default=list,
)
# When True, revlogs use a special reference version of the nodemap, that is not
# performant but is "known" to behave properly.
coreconfigitem(
    b'devel',
    b'persistent-nodemap',
    default=False,
)
coreconfigitem(
    b'devel',
    b'servercafile',
    default=b'',
)
coreconfigitem(
    b'devel',
    b'serverexactprotocol',
    default=b'',
)
coreconfigitem(
    b'devel',
    b'serverrequirecert',
    default=False,
)
coreconfigitem(
    b'devel',
    b'strip-obsmarkers',
    default=True,
)
coreconfigitem(
    b'devel',
    b'warn-config',
    default=None,
)
coreconfigitem(
    b'devel',
    b'warn-config-default',
    default=None,
)
coreconfigitem(
    b'devel',
    b'user.obsmarker',
    default=None,
)
coreconfigitem(
    b'devel',
    b'warn-config-unknown',
    default=None,
)
coreconfigitem(
    b'devel',
    b'debug.copies',
    default=False,
)
coreconfigitem(
    b'devel',
    b'copy-tracing.multi-thread',
    default=True,
)
coreconfigitem(
    b'devel',
    b'debug.extensions',
    default=False,
)
coreconfigitem(
    b'devel',
    b'debug.repo-filters',
    default=False,
)
coreconfigitem(
    b'devel',
    b'debug.peer-request',
    default=False,
)
# If discovery.exchange-heads is False, the discovery will not start with
# remote head fetching and local head querying.
coreconfigitem(
    b'devel',
    b'discovery.exchange-heads',
    default=True,
)
# If discovery.grow-sample is False, the sample size used in set discovery will
# not be increased through the process
coreconfigitem(
    b'devel',
    b'discovery.grow-sample',
    default=True,
)
# When discovery.grow-sample.dynamic is True, the default, the sample size is
# adapted to the shape of the undecided set (it is set to the max of:
# <target-size>, len(roots(undecided)), len(heads(undecided)
coreconfigitem(
    b'devel',
    b'discovery.grow-sample.dynamic',
    default=True,
)
# discovery.grow-sample.rate control the rate at which the sample grow
coreconfigitem(
    b'devel',
    b'discovery.grow-sample.rate',
    default=1.05,
)
# If discovery.randomize is False, random sampling during discovery are
# deterministic. It is meant for integration tests.
coreconfigitem(
    b'devel',
    b'discovery.randomize',
    default=True,
)
# Control the initial size of the discovery sample
coreconfigitem(
    b'devel',
    b'discovery.sample-size',
    default=200,
)
# Control the initial size of the discovery for initial change
coreconfigitem(
    b'devel',
    b'discovery.sample-size.initial',
    default=100,
)
_registerdiffopts(section=b'diff')
coreconfigitem(
    b'diff',
    b'merge',
    default=False,
    experimental=True,
)
coreconfigitem(
    b'email',
    b'bcc',
    default=None,
)
coreconfigitem(
    b'email',
    b'cc',
    default=None,
)
coreconfigitem(
    b'email',
    b'charsets',
    default=list,
)
coreconfigitem(
    b'email',
    b'from',
    default=None,
)
coreconfigitem(
    b'email',
    b'method',
    default=b'smtp',
)
coreconfigitem(
    b'email',
    b'reply-to',
    default=None,
)
coreconfigitem(
    b'email',
    b'to',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'archivemetatemplate',
    default=dynamicdefault,
)
coreconfigitem(
    b'experimental',
    b'auto-publish',
    default=b'publish',
)
coreconfigitem(
    b'experimental',
    b'bundle-phases',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'bundle2-advertise',
    default=True,
)
coreconfigitem(
    b'experimental',
    b'bundle2-output-capture',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'bundle2.pushback',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'bundle2lazylocking',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'bundlecomplevel',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'bundlecomplevel.bzip2',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'bundlecomplevel.gzip',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'bundlecomplevel.none',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'bundlecomplevel.zstd',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'bundlecompthreads',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'bundlecompthreads.bzip2',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'bundlecompthreads.gzip',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'bundlecompthreads.none',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'bundlecompthreads.zstd',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'changegroup3',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'changegroup4',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'cleanup-as-archived',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'clientcompressionengines',
    default=list,
)
coreconfigitem(
    b'experimental',
    b'copytrace',
    default=b'on',
)
coreconfigitem(
    b'experimental',
    b'copytrace.movecandidateslimit',
    default=100,
)
coreconfigitem(
    b'experimental',
    b'copytrace.sourcecommitlimit',
    default=100,
)
coreconfigitem(
    b'experimental',
    b'copies.read-from',
    default=b"filelog-only",
)
coreconfigitem(
    b'experimental',
    b'copies.write-to',
    default=b'filelog-only',
)
coreconfigitem(
    b'experimental',
    b'crecordtest',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'directaccess',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'directaccess.revnums',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'dirstate-tree.in-memory',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'editortmpinhg',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'evolution',
    default=list,
)
coreconfigitem(
    b'experimental',
    b'evolution.allowdivergence',
    default=False,
    alias=[(b'experimental', b'allowdivergence')],
)
coreconfigitem(
    b'experimental',
    b'evolution.allowunstable',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'evolution.createmarkers',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'evolution.effect-flags',
    default=True,
    alias=[(b'experimental', b'effect-flags')],
)
coreconfigitem(
    b'experimental',
    b'evolution.exchange',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'evolution.bundle-obsmarker',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'evolution.bundle-obsmarker:mandatory',
    default=True,
)
coreconfigitem(
    b'experimental',
    b'log.topo',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'evolution.report-instabilities',
    default=True,
)
coreconfigitem(
    b'experimental',
    b'evolution.track-operation',
    default=True,
)
# repo-level config to exclude a revset visibility
#
# The target use case is to use `share` to expose different subset of the same
# repository, especially server side. See also `server.view`.
coreconfigitem(
    b'experimental',
    b'extra-filter-revs',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'maxdeltachainspan',
    default=-1,
)
# tracks files which were undeleted (merge might delete them but we explicitly
# kept/undeleted them) and creates new filenodes for them
coreconfigitem(
    b'experimental',
    b'merge-track-salvaged',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'mergetempdirprefix',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'mmapindexthreshold',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'narrow',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'nonnormalparanoidcheck',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'exportableenviron',
    default=list,
)
coreconfigitem(
    b'experimental',
    b'extendedheader.index',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'extendedheader.similarity',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'graphshorten',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'graphstyle.parent',
    default=dynamicdefault,
)
coreconfigitem(
    b'experimental',
    b'graphstyle.missing',
    default=dynamicdefault,
)
coreconfigitem(
    b'experimental',
    b'graphstyle.grandparent',
    default=dynamicdefault,
)
coreconfigitem(
    b'experimental',
    b'hook-track-tags',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'httppeer.advertise-v2',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'httppeer.v2-encoder-order',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'httppostargs',
    default=False,
)
coreconfigitem(b'experimental', b'nointerrupt', default=False)
coreconfigitem(b'experimental', b'nointerrupt-interactiveonly', default=True)

coreconfigitem(
    b'experimental',
    b'obsmarkers-exchange-debug',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'remotenames',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'removeemptydirs',
    default=True,
)
coreconfigitem(
    b'experimental',
    b'revert.interactive.select-to-keep',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'revisions.prefixhexnode',
    default=False,
)
# "out of experimental" todo list.
#
# * include management of a persistent nodemap in the main docket
# * enforce a "no-truncate" policy for mmap safety
#      - for censoring operation
#      - for stripping operation
#      - for rollback operation
# * proper streaming (race free) of the docket file
# * track garbage data to evemtually allow rewriting -existing- sidedata.
# * Exchange-wise, we will also need to do something more efficient than
#   keeping references to the affected revlogs, especially memory-wise when
#   rewriting sidedata.
# * introduce a proper solution to reduce the number of filelog related files.
# * use caching for reading sidedata (similar to what we do for data).
# * no longer set offset=0 if sidedata_size=0 (simplify cutoff computation).
# * Improvement to consider
#   - avoid compression header in chunk using the default compression?
#   - forbid "inline" compression mode entirely?
#   - split the data offset and flag field (the 2 bytes save are mostly trouble)
#   - keep track of uncompressed -chunk- size (to preallocate memory better)
#   - keep track of chain base or size (probably not that useful anymore)
coreconfigitem(
    b'experimental',
    b'revlogv2',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'revisions.disambiguatewithin',
    default=None,
)
coreconfigitem(
    b'experimental',
    b'rust.index',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'server.filesdata.recommended-batch-size',
    default=50000,
)
coreconfigitem(
    b'experimental',
    b'server.manifestdata.recommended-batch-size',
    default=100000,
)
coreconfigitem(
    b'experimental',
    b'server.stream-narrow-clones',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'single-head-per-branch',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'single-head-per-branch:account-closed-heads',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'single-head-per-branch:public-changes-only',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'sshserver.support-v2',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'sparse-read',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'sparse-read.density-threshold',
    default=0.50,
)
coreconfigitem(
    b'experimental',
    b'sparse-read.min-gap-size',
    default=b'65K',
)
coreconfigitem(
    b'experimental',
    b'treemanifest',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'update.atomic-file',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'sshpeer.advertise-v2',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'web.apiserver',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'web.api.http-v2',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'web.api.debugreflect',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'worker.wdir-get-thread-safe',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'worker.repository-upgrade',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'xdiff',
    default=False,
)
coreconfigitem(
    b'extensions',
    b'.*',
    default=None,
    generic=True,
)
coreconfigitem(
    b'extdata',
    b'.*',
    default=None,
    generic=True,
)
coreconfigitem(
    b'format',
    b'bookmarks-in-store',
    default=False,
)
coreconfigitem(
    b'format',
    b'chunkcachesize',
    default=None,
    experimental=True,
)
coreconfigitem(
    # Enable this dirstate format *when creating a new repository*.
    # Which format to use for existing repos is controlled by .hg/requires
    b'format',
    b'exp-dirstate-v2',
    default=False,
    experimental=True,
)
coreconfigitem(
    b'format',
    b'dotencode',
    default=True,
)
coreconfigitem(
    b'format',
    b'generaldelta',
    default=False,
    experimental=True,
)
coreconfigitem(
    b'format',
    b'manifestcachesize',
    default=None,
    experimental=True,
)
coreconfigitem(
    b'format',
    b'maxchainlen',
    default=dynamicdefault,
    experimental=True,
)
coreconfigitem(
    b'format',
    b'obsstore-version',
    default=None,
)
coreconfigitem(
    b'format',
    b'sparse-revlog',
    default=True,
)
coreconfigitem(
    b'format',
    b'revlog-compression',
    default=lambda: [b'zstd', b'zlib'],
    alias=[(b'experimental', b'format.compression')],
)
# Experimental TODOs:
#
# * Same as for evlogv2 (but for the reduction of the number of files)
# * Improvement to investigate
#   - storing .hgtags fnode
#   - storing `rank` of changesets
#   - storing branch related identifier

coreconfigitem(
    b'format',
    b'exp-use-changelog-v2',
    default=None,
    experimental=True,
)
coreconfigitem(
    b'format',
    b'usefncache',
    default=True,
)
coreconfigitem(
    b'format',
    b'usegeneraldelta',
    default=True,
)
coreconfigitem(
    b'format',
    b'usestore',
    default=True,
)


def _persistent_nodemap_default():
    """compute `use-persistent-nodemap` default value

    The feature is disabled unless a fast implementation is available.
    """
    from . import policy

    return policy.importrust('revlog') is not None


coreconfigitem(
    b'format',
    b'use-persistent-nodemap',
    default=_persistent_nodemap_default,
)
coreconfigitem(
    b'format',
    b'exp-use-copies-side-data-changeset',
    default=False,
    experimental=True,
)
coreconfigitem(
    b'format',
    b'use-share-safe',
    default=False,
)
coreconfigitem(
    b'format',
    b'internal-phase',
    default=False,
    experimental=True,
)
coreconfigitem(
    b'fsmonitor',
    b'warn_when_unused',
    default=True,
)
coreconfigitem(
    b'fsmonitor',
    b'warn_update_file_count',
    default=50000,
)
coreconfigitem(
    b'fsmonitor',
    b'warn_update_file_count_rust',
    default=400000,
)
coreconfigitem(
    b'help',
    br'hidden-command\..*',
    default=False,
    generic=True,
)
coreconfigitem(
    b'help',
    br'hidden-topic\..*',
    default=False,
    generic=True,
)
coreconfigitem(
    b'hooks',
    b'[^:]*',
    default=dynamicdefault,
    generic=True,
)
coreconfigitem(
    b'hooks',
    b'.*:run-with-plain',
    default=True,
    generic=True,
)
coreconfigitem(
    b'hgweb-paths',
    b'.*',
    default=list,
    generic=True,
)
coreconfigitem(
    b'hostfingerprints',
    b'.*',
    default=list,
    generic=True,
)
coreconfigitem(
    b'hostsecurity',
    b'ciphers',
    default=None,
)
coreconfigitem(
    b'hostsecurity',
    b'minimumprotocol',
    default=dynamicdefault,
)
coreconfigitem(
    b'hostsecurity',
    b'.*:minimumprotocol$',
    default=dynamicdefault,
    generic=True,
)
coreconfigitem(
    b'hostsecurity',
    b'.*:ciphers$',
    default=dynamicdefault,
    generic=True,
)
coreconfigitem(
    b'hostsecurity',
    b'.*:fingerprints$',
    default=list,
    generic=True,
)
coreconfigitem(
    b'hostsecurity',
    b'.*:verifycertsfile$',
    default=None,
    generic=True,
)

coreconfigitem(
    b'http_proxy',
    b'always',
    default=False,
)
coreconfigitem(
    b'http_proxy',
    b'host',
    default=None,
)
coreconfigitem(
    b'http_proxy',
    b'no',
    default=list,
)
coreconfigitem(
    b'http_proxy',
    b'passwd',
    default=None,
)
coreconfigitem(
    b'http_proxy',
    b'user',
    default=None,
)

coreconfigitem(
    b'http',
    b'timeout',
    default=None,
)

coreconfigitem(
    b'logtoprocess',
    b'commandexception',
    default=None,
)
coreconfigitem(
    b'logtoprocess',
    b'commandfinish',
    default=None,
)
coreconfigitem(
    b'logtoprocess',
    b'command',
    default=None,
)
coreconfigitem(
    b'logtoprocess',
    b'develwarn',
    default=None,
)
coreconfigitem(
    b'logtoprocess',
    b'uiblocked',
    default=None,
)
coreconfigitem(
    b'merge',
    b'checkunknown',
    default=b'abort',
)
coreconfigitem(
    b'merge',
    b'checkignored',
    default=b'abort',
)
coreconfigitem(
    b'experimental',
    b'merge.checkpathconflicts',
    default=False,
)
coreconfigitem(
    b'merge',
    b'followcopies',
    default=True,
)
coreconfigitem(
    b'merge',
    b'on-failure',
    default=b'continue',
)
coreconfigitem(
    b'merge',
    b'preferancestor',
    default=lambda: [b'*'],
    experimental=True,
)
coreconfigitem(
    b'merge',
    b'strict-capability-check',
    default=False,
)
coreconfigitem(
    b'merge-tools',
    b'.*',
    default=None,
    generic=True,
)
coreconfigitem(
    b'merge-tools',
    br'.*\.args$',
    default=b"$local $base $other",
    generic=True,
    priority=-1,
)
coreconfigitem(
    b'merge-tools',
    br'.*\.binary$',
    default=False,
    generic=True,
    priority=-1,
)
coreconfigitem(
    b'merge-tools',
    br'.*\.check$',
    default=list,
    generic=True,
    priority=-1,
)
coreconfigitem(
    b'merge-tools',
    br'.*\.checkchanged$',
    default=False,
    generic=True,
    priority=-1,
)
coreconfigitem(
    b'merge-tools',
    br'.*\.executable$',
    default=dynamicdefault,
    generic=True,
    priority=-1,
)
coreconfigitem(
    b'merge-tools',
    br'.*\.fixeol$',
    default=False,
    generic=True,
    priority=-1,
)
coreconfigitem(
    b'merge-tools',
    br'.*\.gui$',
    default=False,
    generic=True,
    priority=-1,
)
coreconfigitem(
    b'merge-tools',
    br'.*\.mergemarkers$',
    default=b'basic',
    generic=True,
    priority=-1,
)
coreconfigitem(
    b'merge-tools',
    br'.*\.mergemarkertemplate$',
    default=dynamicdefault,  # take from command-templates.mergemarker
    generic=True,
    priority=-1,
)
coreconfigitem(
    b'merge-tools',
    br'.*\.priority$',
    default=0,
    generic=True,
    priority=-1,
)
coreconfigitem(
    b'merge-tools',
    br'.*\.premerge$',
    default=dynamicdefault,
    generic=True,
    priority=-1,
)
coreconfigitem(
    b'merge-tools',
    br'.*\.symlink$',
    default=False,
    generic=True,
    priority=-1,
)
coreconfigitem(
    b'pager',
    b'attend-.*',
    default=dynamicdefault,
    generic=True,
)
coreconfigitem(
    b'pager',
    b'ignore',
    default=list,
)
coreconfigitem(
    b'pager',
    b'pager',
    default=dynamicdefault,
)
coreconfigitem(
    b'patch',
    b'eol',
    default=b'strict',
)
coreconfigitem(
    b'patch',
    b'fuzz',
    default=2,
)
coreconfigitem(
    b'paths',
    b'default',
    default=None,
)
coreconfigitem(
    b'paths',
    b'default-push',
    default=None,
)
coreconfigitem(
    b'paths',
    b'.*',
    default=None,
    generic=True,
)
coreconfigitem(
    b'phases',
    b'checksubrepos',
    default=b'follow',
)
coreconfigitem(
    b'phases',
    b'new-commit',
    default=b'draft',
)
coreconfigitem(
    b'phases',
    b'publish',
    default=True,
)
coreconfigitem(
    b'profiling',
    b'enabled',
    default=False,
)
coreconfigitem(
    b'profiling',
    b'format',
    default=b'text',
)
coreconfigitem(
    b'profiling',
    b'freq',
    default=1000,
)
coreconfigitem(
    b'profiling',
    b'limit',
    default=30,
)
coreconfigitem(
    b'profiling',
    b'nested',
    default=0,
)
coreconfigitem(
    b'profiling',
    b'output',
    default=None,
)
coreconfigitem(
    b'profiling',
    b'showmax',
    default=0.999,
)
coreconfigitem(
    b'profiling',
    b'showmin',
    default=dynamicdefault,
)
coreconfigitem(
    b'profiling',
    b'showtime',
    default=True,
)
coreconfigitem(
    b'profiling',
    b'sort',
    default=b'inlinetime',
)
coreconfigitem(
    b'profiling',
    b'statformat',
    default=b'hotpath',
)
coreconfigitem(
    b'profiling',
    b'time-track',
    default=dynamicdefault,
)
coreconfigitem(
    b'profiling',
    b'type',
    default=b'stat',
)
coreconfigitem(
    b'progress',
    b'assume-tty',
    default=False,
)
coreconfigitem(
    b'progress',
    b'changedelay',
    default=1,
)
coreconfigitem(
    b'progress',
    b'clear-complete',
    default=True,
)
coreconfigitem(
    b'progress',
    b'debug',
    default=False,
)
coreconfigitem(
    b'progress',
    b'delay',
    default=3,
)
coreconfigitem(
    b'progress',
    b'disable',
    default=False,
)
coreconfigitem(
    b'progress',
    b'estimateinterval',
    default=60.0,
)
coreconfigitem(
    b'progress',
    b'format',
    default=lambda: [b'topic', b'bar', b'number', b'estimate'],
)
coreconfigitem(
    b'progress',
    b'refresh',
    default=0.1,
)
coreconfigitem(
    b'progress',
    b'width',
    default=dynamicdefault,
)
coreconfigitem(
    b'pull',
    b'confirm',
    default=False,
)
coreconfigitem(
    b'push',
    b'pushvars.server',
    default=False,
)
coreconfigitem(
    b'rewrite',
    b'backup-bundle',
    default=True,
    alias=[(b'ui', b'history-editing-backup')],
)
coreconfigitem(
    b'rewrite',
    b'update-timestamp',
    default=False,
)
coreconfigitem(
    b'rewrite',
    b'empty-successor',
    default=b'skip',
    experimental=True,
)
coreconfigitem(
    b'storage',
    b'new-repo-backend',
    default=b'revlogv1',
    experimental=True,
)
coreconfigitem(
    b'storage',
    b'revlog.optimize-delta-parent-choice',
    default=True,
    alias=[(b'format', b'aggressivemergedeltas')],
)
coreconfigitem(
    b'storage',
    b'revlog.issue6528.fix-incoming',
    default=True,
)
# experimental as long as rust is experimental (or a C version is implemented)
coreconfigitem(
    b'storage',
    b'revlog.persistent-nodemap.mmap',
    default=True,
)
# experimental as long as format.use-persistent-nodemap is.
coreconfigitem(
    b'storage',
    b'revlog.persistent-nodemap.slow-path',
    default=b"abort",
)

coreconfigitem(
    b'storage',
    b'revlog.reuse-external-delta',
    default=True,
)
coreconfigitem(
    b'storage',
    b'revlog.reuse-external-delta-parent',
    default=None,
)
coreconfigitem(
    b'storage',
    b'revlog.zlib.level',
    default=None,
)
coreconfigitem(
    b'storage',
    b'revlog.zstd.level',
    default=None,
)
coreconfigitem(
    b'server',
    b'bookmarks-pushkey-compat',
    default=True,
)
coreconfigitem(
    b'server',
    b'bundle1',
    default=True,
)
coreconfigitem(
    b'server',
    b'bundle1gd',
    default=None,
)
coreconfigitem(
    b'server',
    b'bundle1.pull',
    default=None,
)
coreconfigitem(
    b'server',
    b'bundle1gd.pull',
    default=None,
)
coreconfigitem(
    b'server',
    b'bundle1.push',
    default=None,
)
coreconfigitem(
    b'server',
    b'bundle1gd.push',
    default=None,
)
coreconfigitem(
    b'server',
    b'bundle2.stream',
    default=True,
    alias=[(b'experimental', b'bundle2.stream')],
)
coreconfigitem(
    b'server',
    b'compressionengines',
    default=list,
)
coreconfigitem(
    b'server',
    b'concurrent-push-mode',
    default=b'check-related',
)
coreconfigitem(
    b'server',
    b'disablefullbundle',
    default=False,
)
coreconfigitem(
    b'server',
    b'maxhttpheaderlen',
    default=1024,
)
coreconfigitem(
    b'server',
    b'pullbundle',
    default=False,
)
coreconfigitem(
    b'server',
    b'preferuncompressed',
    default=False,
)
coreconfigitem(
    b'server',
    b'streamunbundle',
    default=False,
)
coreconfigitem(
    b'server',
    b'uncompressed',
    default=True,
)
coreconfigitem(
    b'server',
    b'uncompressedallowsecret',
    default=False,
)
coreconfigitem(
    b'server',
    b'view',
    default=b'served',
)
coreconfigitem(
    b'server',
    b'validate',
    default=False,
)
coreconfigitem(
    b'server',
    b'zliblevel',
    default=-1,
)
coreconfigitem(
    b'server',
    b'zstdlevel',
    default=3,
)
coreconfigitem(
    b'share',
    b'pool',
    default=None,
)
coreconfigitem(
    b'share',
    b'poolnaming',
    default=b'identity',
)
coreconfigitem(
    b'share',
    b'safe-mismatch.source-not-safe',
    default=b'abort',
)
coreconfigitem(
    b'share',
    b'safe-mismatch.source-safe',
    default=b'abort',
)
coreconfigitem(
    b'share',
    b'safe-mismatch.source-not-safe.warn',
    default=True,
)
coreconfigitem(
    b'share',
    b'safe-mismatch.source-safe.warn',
    default=True,
)
coreconfigitem(
    b'shelve',
    b'maxbackups',
    default=10,
)
coreconfigitem(
    b'smtp',
    b'host',
    default=None,
)
coreconfigitem(
    b'smtp',
    b'local_hostname',
    default=None,
)
coreconfigitem(
    b'smtp',
    b'password',
    default=None,
)
coreconfigitem(
    b'smtp',
    b'port',
    default=dynamicdefault,
)
coreconfigitem(
    b'smtp',
    b'tls',
    default=b'none',
)
coreconfigitem(
    b'smtp',
    b'username',
    default=None,
)
coreconfigitem(
    b'sparse',
    b'missingwarning',
    default=True,
    experimental=True,
)
coreconfigitem(
    b'subrepos',
    b'allowed',
    default=dynamicdefault,  # to make backporting simpler
)
coreconfigitem(
    b'subrepos',
    b'hg:allowed',
    default=dynamicdefault,
)
coreconfigitem(
    b'subrepos',
    b'git:allowed',
    default=dynamicdefault,
)
coreconfigitem(
    b'subrepos',
    b'svn:allowed',
    default=dynamicdefault,
)
coreconfigitem(
    b'templates',
    b'.*',
    default=None,
    generic=True,
)
coreconfigitem(
    b'templateconfig',
    b'.*',
    default=dynamicdefault,
    generic=True,
)
coreconfigitem(
    b'trusted',
    b'groups',
    default=list,
)
coreconfigitem(
    b'trusted',
    b'users',
    default=list,
)
coreconfigitem(
    b'ui',
    b'_usedassubrepo',
    default=False,
)
coreconfigitem(
    b'ui',
    b'allowemptycommit',
    default=False,
)
coreconfigitem(
    b'ui',
    b'archivemeta',
    default=True,
)
coreconfigitem(
    b'ui',
    b'askusername',
    default=False,
)
coreconfigitem(
    b'ui',
    b'available-memory',
    default=None,
)

coreconfigitem(
    b'ui',
    b'clonebundlefallback',
    default=False,
)
coreconfigitem(
    b'ui',
    b'clonebundleprefers',
    default=list,
)
coreconfigitem(
    b'ui',
    b'clonebundles',
    default=True,
)
coreconfigitem(
    b'ui',
    b'color',
    default=b'auto',
)
coreconfigitem(
    b'ui',
    b'commitsubrepos',
    default=False,
)
coreconfigitem(
    b'ui',
    b'debug',
    default=False,
)
coreconfigitem(
    b'ui',
    b'debugger',
    default=None,
)
coreconfigitem(
    b'ui',
    b'editor',
    default=dynamicdefault,
)
coreconfigitem(
    b'ui',
    b'detailed-exit-code',
    default=False,
    experimental=True,
)
coreconfigitem(
    b'ui',
    b'fallbackencoding',
    default=None,
)
coreconfigitem(
    b'ui',
    b'forcecwd',
    default=None,
)
coreconfigitem(
    b'ui',
    b'forcemerge',
    default=None,
)
coreconfigitem(
    b'ui',
    b'formatdebug',
    default=False,
)
coreconfigitem(
    b'ui',
    b'formatjson',
    default=False,
)
coreconfigitem(
    b'ui',
    b'formatted',
    default=None,
)
coreconfigitem(
    b'ui',
    b'interactive',
    default=None,
)
coreconfigitem(
    b'ui',
    b'interface',
    default=None,
)
coreconfigitem(
    b'ui',
    b'interface.chunkselector',
    default=None,
)
coreconfigitem(
    b'ui',
    b'large-file-limit',
    default=10000000,
)
coreconfigitem(
    b'ui',
    b'logblockedtimes',
    default=False,
)
coreconfigitem(
    b'ui',
    b'merge',
    default=None,
)
coreconfigitem(
    b'ui',
    b'mergemarkers',
    default=b'basic',
)
coreconfigitem(
    b'ui',
    b'message-output',
    default=b'stdio',
)
coreconfigitem(
    b'ui',
    b'nontty',
    default=False,
)
coreconfigitem(
    b'ui',
    b'origbackuppath',
    default=None,
)
coreconfigitem(
    b'ui',
    b'paginate',
    default=True,
)
coreconfigitem(
    b'ui',
    b'patch',
    default=None,
)
coreconfigitem(
    b'ui',
    b'portablefilenames',
    default=b'warn',
)
coreconfigitem(
    b'ui',
    b'promptecho',
    default=False,
)
coreconfigitem(
    b'ui',
    b'quiet',
    default=False,
)
coreconfigitem(
    b'ui',
    b'quietbookmarkmove',
    default=False,
)
coreconfigitem(
    b'ui',
    b'relative-paths',
    default=b'legacy',
)
coreconfigitem(
    b'ui',
    b'remotecmd',
    default=b'hg',
)
coreconfigitem(
    b'ui',
    b'report_untrusted',
    default=True,
)
coreconfigitem(
    b'ui',
    b'rollback',
    default=True,
)
coreconfigitem(
    b'ui',
    b'signal-safe-lock',
    default=True,
)
coreconfigitem(
    b'ui',
    b'slash',
    default=False,
)
coreconfigitem(
    b'ui',
    b'ssh',
    default=b'ssh',
)
coreconfigitem(
    b'ui',
    b'ssherrorhint',
    default=None,
)
coreconfigitem(
    b'ui',
    b'statuscopies',
    default=False,
)
coreconfigitem(
    b'ui',
    b'strict',
    default=False,
)
coreconfigitem(
    b'ui',
    b'style',
    default=b'',
)
coreconfigitem(
    b'ui',
    b'supportcontact',
    default=None,
)
coreconfigitem(
    b'ui',
    b'textwidth',
    default=78,
)
coreconfigitem(
    b'ui',
    b'timeout',
    default=b'600',
)
coreconfigitem(
    b'ui',
    b'timeout.warn',
    default=0,
)
coreconfigitem(
    b'ui',
    b'timestamp-output',
    default=False,
)
coreconfigitem(
    b'ui',
    b'traceback',
    default=False,
)
coreconfigitem(
    b'ui',
    b'tweakdefaults',
    default=False,
)
coreconfigitem(b'ui', b'username', alias=[(b'ui', b'user')])
coreconfigitem(
    b'ui',
    b'verbose',
    default=False,
)
coreconfigitem(
    b'verify',
    b'skipflags',
    default=None,
)
coreconfigitem(
    b'web',
    b'allowbz2',
    default=False,
)
coreconfigitem(
    b'web',
    b'allowgz',
    default=False,
)
coreconfigitem(
    b'web',
    b'allow-pull',
    alias=[(b'web', b'allowpull')],
    default=True,
)
coreconfigitem(
    b'web',
    b'allow-push',
    alias=[(b'web', b'allow_push')],
    default=list,
)
coreconfigitem(
    b'web',
    b'allowzip',
    default=False,
)
coreconfigitem(
    b'web',
    b'archivesubrepos',
    default=False,
)
coreconfigitem(
    b'web',
    b'cache',
    default=True,
)
coreconfigitem(
    b'web',
    b'comparisoncontext',
    default=5,
)
coreconfigitem(
    b'web',
    b'contact',
    default=None,
)
coreconfigitem(
    b'web',
    b'deny_push',
    default=list,
)
coreconfigitem(
    b'web',
    b'guessmime',
    default=False,
)
coreconfigitem(
    b'web',
    b'hidden',
    default=False,
)
coreconfigitem(
    b'web',
    b'labels',
    default=list,
)
coreconfigitem(
    b'web',
    b'logoimg',
    default=b'hglogo.png',
)
coreconfigitem(
    b'web',
    b'logourl',
    default=b'https://mercurial-scm.org/',
)
coreconfigitem(
    b'web',
    b'accesslog',
    default=b'-',
)
coreconfigitem(
    b'web',
    b'address',
    default=b'',
)
coreconfigitem(
    b'web',
    b'allow-archive',
    alias=[(b'web', b'allow_archive')],
    default=list,
)
coreconfigitem(
    b'web',
    b'allow_read',
    default=list,
)
coreconfigitem(
    b'web',
    b'baseurl',
    default=None,
)
coreconfigitem(
    b'web',
    b'cacerts',
    default=None,
)
coreconfigitem(
    b'web',
    b'certificate',
    default=None,
)
coreconfigitem(
    b'web',
    b'collapse',
    default=False,
)
coreconfigitem(
    b'web',
    b'csp',
    default=None,
)
coreconfigitem(
    b'web',
    b'deny_read',
    default=list,
)
coreconfigitem(
    b'web',
    b'descend',
    default=True,
)
coreconfigitem(
    b'web',
    b'description',
    default=b"",
)
coreconfigitem(
    b'web',
    b'encoding',
    default=lambda: encoding.encoding,
)
coreconfigitem(
    b'web',
    b'errorlog',
    default=b'-',
)
coreconfigitem(
    b'web',
    b'ipv6',
    default=False,
)
coreconfigitem(
    b'web',
    b'maxchanges',
    default=10,
)
coreconfigitem(
    b'web',
    b'maxfiles',
    default=10,
)
coreconfigitem(
    b'web',
    b'maxshortchanges',
    default=60,
)
coreconfigitem(
    b'web',
    b'motd',
    default=b'',
)
coreconfigitem(
    b'web',
    b'name',
    default=dynamicdefault,
)
coreconfigitem(
    b'web',
    b'port',
    default=8000,
)
coreconfigitem(
    b'web',
    b'prefix',
    default=b'',
)
coreconfigitem(
    b'web',
    b'push_ssl',
    default=True,
)
coreconfigitem(
    b'web',
    b'refreshinterval',
    default=20,
)
coreconfigitem(
    b'web',
    b'server-header',
    default=None,
)
coreconfigitem(
    b'web',
    b'static',
    default=None,
)
coreconfigitem(
    b'web',
    b'staticurl',
    default=None,
)
coreconfigitem(
    b'web',
    b'stripes',
    default=1,
)
coreconfigitem(
    b'web',
    b'style',
    default=b'paper',
)
coreconfigitem(
    b'web',
    b'templates',
    default=None,
)
coreconfigitem(
    b'web',
    b'view',
    default=b'served',
    experimental=True,
)
coreconfigitem(
    b'worker',
    b'backgroundclose',
    default=dynamicdefault,
)
# Windows defaults to a limit of 512 open files. A buffer of 128
# should give us enough headway.
coreconfigitem(
    b'worker',
    b'backgroundclosemaxqueue',
    default=384,
)
coreconfigitem(
    b'worker',
    b'backgroundcloseminfilecount',
    default=2048,
)
coreconfigitem(
    b'worker',
    b'backgroundclosethreadcount',
    default=4,
)
coreconfigitem(
    b'worker',
    b'enabled',
    default=True,
)
coreconfigitem(
    b'worker',
    b'numcpus',
    default=None,
)

# Rebase related configuration moved to core because other extension are doing
# strange things. For example, shelve import the extensions to reuse some bit
# without formally loading it.
coreconfigitem(
    b'commands',
    b'rebase.requiredest',
    default=False,
)
coreconfigitem(
    b'experimental',
    b'rebaseskipobsolete',
    default=True,
)
coreconfigitem(
    b'rebase',
    b'singletransaction',
    default=False,
)
coreconfigitem(
    b'rebase',
    b'experimental.inmemory',
    default=False,
)



# streamclone.py - producing and consuming streaming repository data
#
# Copyright 2015 Gregory Szorc <gregory.szorc@gmail.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import absolute_import

import contextlib
import errno
import os
import struct

from .i18n import _
from .pycompat import open
from .interfaces import repository
from . import (
    bookmarks,
    cacheutil,
    error,
    narrowspec,
    phases,
    pycompat,
    requirements as requirementsmod,
    scmutil,
    store,
    util,
)
from .utils import (
    stringutil,
)


def canperformstreamclone(pullop, bundle2=False):
    """Whether it is possible to perform a streaming clone as part of pull.

    ``bundle2`` will cause the function to consider stream clone through
    bundle2 and only through bundle2.

    Returns a tuple of (supported, requirements). ``supported`` is True if
    streaming clone is supported and False otherwise. ``requirements`` is
    a set of repo requirements from the remote, or ``None`` if stream clone
    isn't supported.
    """
    repo = pullop.repo
    remote = pullop.remote

    bundle2supported = False
    if pullop.canusebundle2:
        if b'v2' in pullop.remotebundle2caps.get(b'stream', []):
            bundle2supported = True
        # else
        # Server doesn't support bundle2 stream clone or doesn't support
        # the versions we support. Fall back and possibly allow legacy.

    # Ensures legacy code path uses available bundle2.
    if bundle2supported and not bundle2:
        return False, None
    # Ensures bundle2 doesn't try to do a stream clone if it isn't supported.
    elif bundle2 and not bundle2supported:
        return False, None

    # Streaming clone only works on empty repositories.
    if len(repo):
        return False, None

    # Streaming clone only works if all data is being requested.
    if pullop.heads:
        return False, None

    streamrequested = pullop.streamclonerequested

    # If we don't have a preference, let the server decide for us. This
    # likely only comes into play in LANs.
    if streamrequested is None:
        # The server can advertise whether to prefer streaming clone.
        streamrequested = remote.capable(b'stream-preferred')

    if not streamrequested:
        return False, None

    # In order for stream clone to work, the client has to support all the
    # requirements advertised by the server.
    #
    # The server advertises its requirements via the "stream" and "streamreqs"
    # capability. "stream" (a value-less capability) is advertised if and only
    # if the only requirement is "revlogv1." Else, the "streamreqs" capability
    # is advertised and contains a comma-delimited list of requirements.
    requirements = set()
    if remote.capable(b'stream'):
        requirements.add(requirementsmod.REVLOGV1_REQUIREMENT)
    else:
        streamreqs = remote.capable(b'streamreqs')
        # This is weird and shouldn't happen with modern servers.
        if not streamreqs:
            pullop.repo.ui.warn(
                _(
                    b'warning: stream clone requested but server has them '
                    b'disabled\n'
                )
            )
            return False, None

        streamreqs = set(streamreqs.split(b','))
        # Server requires something we don't support. Bail.
        missingreqs = streamreqs - repo.supportedformats
        if missingreqs:
            pullop.repo.ui.warn(
                _(
                    b'warning: stream clone requested but client is missing '
                    b'requirements: %s\n'
                )
                % b', '.join(sorted(missingreqs))
            )
            pullop.repo.ui.warn(
                _(
                    b'(see https://www.mercurial-scm.org/wiki/MissingRequirement '
                    b'for more information)\n'
                )
            )
            return False, None
        requirements = streamreqs

    return True, requirements


def maybeperformlegacystreamclone(pullop):
    """Possibly perform a legacy stream clone operation.

    Legacy stream clones are performed as part of pull but before all other
    operations.

    A legacy stream clone will not be performed if a bundle2 stream clone is
    supported.
    """
    from . import localrepo

    supported, requirements = canperformstreamclone(pullop)

    if not supported:
        return

    repo = pullop.repo
    remote = pullop.remote

    # Save remote branchmap. We will use it later to speed up branchcache
    # creation.
    rbranchmap = None
    if remote.capable(b'branchmap'):
        with remote.commandexecutor() as e:
            rbranchmap = e.callcommand(b'branchmap', {}).result()

    repo.ui.status(_(b'streaming all changes\n'))

    with remote.commandexecutor() as e:
        fp = e.callcommand(b'stream_out', {}).result()

    # TODO strictly speaking, this code should all be inside the context
    # manager because the context manager is supposed to ensure all wire state
    # is flushed when exiting. But the legacy peers don't do this, so it
    # doesn't matter.
    l = fp.readline()
    try:
        resp = int(l)
    except ValueError:
        raise error.ResponseError(
            _(b'unexpected response from remote server:'), l
        )
    if resp == 1:
        raise error.Abort(_(b'operation forbidden by server'))
    elif resp == 2:
        raise error.Abort(_(b'locking the remote repository failed'))
    elif resp != 0:
        raise error.Abort(_(b'the server sent an unknown error code'))

    l = fp.readline()
    try:
        filecount, bytecount = map(int, l.split(b' ', 1))
    except (ValueError, TypeError):
        raise error.ResponseError(
            _(b'unexpected response from remote server:'), l
        )

    with repo.lock():
        consumev1(repo, fp, filecount, bytecount)

        # new requirements = old non-format requirements +
        #                    new format-related remote requirements
        # requirements from the streamed-in repository
        repo.requirements = requirements | (
            repo.requirements - repo.supportedformats
        )
        repo.svfs.options = localrepo.resolvestorevfsoptions(
            repo.ui, repo.requirements, repo.features
        )
        scmutil.writereporequirements(repo)

        if rbranchmap:
            repo._branchcaches.replace(repo, rbranchmap)

        repo.invalidate()


def allowservergeneration(repo):
    """Whether streaming clones are allowed from the server."""
    if repository.REPO_FEATURE_STREAM_CLONE not in repo.features:
        return False

    if not repo.ui.configbool(b'server', b'uncompressed', untrusted=True):
        return False

    # The way stream clone works makes it impossible to hide secret changesets.
    # So don't allow this by default.
    secret = phases.hassecret(repo)
    if secret:
        return repo.ui.configbool(b'server', b'uncompressedallowsecret')

    return True


# This is it's own function so extensions can override it.
def _walkstreamfiles(repo, matcher=None):
    return repo.store.walk(matcher)


def generatev1(repo):
    """Emit content for version 1 of a streaming clone.

    This returns a 3-tuple of (file count, byte size, data iterator).

    The data iterator consists of N entries for each file being transferred.
    Each file entry starts as a line with the file name and integer size
    delimited by a null byte.

    The raw file data follows. Following the raw file data is the next file
    entry, or EOF.

    When used on the wire protocol, an additional line indicating protocol
    success will be prepended to the stream. This function is not responsible
    for adding it.

    This function will obtain a repository lock to ensure a consistent view of
    the store is captured. It therefore may raise LockError.
    """
    entries = []
    total_bytes = 0
    # Get consistent snapshot of repo, lock during scan.
    with repo.lock():
        repo.ui.debug(b'scanning\n')
        for file_type, name, ename, size in _walkstreamfiles(repo):
            if size:
                entries.append((name, size))
                total_bytes += size
        _test_sync_point_walk_1(repo)
    _test_sync_point_walk_2(repo)

    repo.ui.debug(
        b'%d files, %d bytes to transfer\n' % (len(entries), total_bytes)
    )

    svfs = repo.svfs
    debugflag = repo.ui.debugflag

    def emitrevlogdata():
        for name, size in entries:
            if debugflag:
                repo.ui.debug(b'sending %s (%d bytes)\n' % (name, size))
            # partially encode name over the wire for backwards compat
            yield b'%s\0%d\n' % (store.encodedir(name), size)
            # auditing at this stage is both pointless (paths are already
            # trusted by the local repo) and expensive
            with svfs(name, b'rb', auditpath=False) as fp:
                if size <= 65536:
                    yield fp.read(size)
                else:
                    for chunk in util.filechunkiter(fp, limit=size):
                        yield chunk

    return len(entries), total_bytes, emitrevlogdata()


def generatev1wireproto(repo):
    """Emit content for version 1 of streaming clone suitable for the wire.

    This is the data output from ``generatev1()`` with 2 header lines. The
    first line indicates overall success. The 2nd contains the file count and
    byte size of payload.

    The success line contains "0" for success, "1" for stream generation not
    allowed, and "2" for error locking the repository (possibly indicating
    a permissions error for the server process).
    """
    if not allowservergeneration(repo):
        yield b'1\n'
        return

    try:
        filecount, bytecount, it = generatev1(repo)
    except error.LockError:
        yield b'2\n'
        return

    # Indicates successful response.
    yield b'0\n'
    yield b'%d %d\n' % (filecount, bytecount)
    for chunk in it:
        yield chunk


def generatebundlev1(repo, compression=b'UN'):
    """Emit content for version 1 of a stream clone bundle.

    The first 4 bytes of the output ("HGS1") denote this as stream clone
    bundle version 1.

    The next 2 bytes indicate the compression type. Only "UN" is currently
    supported.

    The next 16 bytes are two 64-bit big endian unsigned integers indicating
    file count and byte count, respectively.

    The next 2 bytes is a 16-bit big endian unsigned short declaring the length
    of the requirements string, including a trailing \0. The following N bytes
    are the requirements string, which is ASCII containing a comma-delimited
    list of repo requirements that are needed to support the data.

    The remaining content is the output of ``generatev1()`` (which may be
    compressed in the future).

    Returns a tuple of (requirements, data generator).
    """
    if compression != b'UN':
        raise ValueError(b'we do not support the compression argument yet')

    requirements = repo.requirements & repo.supportedformats
    requires = b','.join(sorted(requirements))

    def gen():
        yield b'HGS1'
        yield compression

        filecount, bytecount, it = generatev1(repo)
        repo.ui.status(
            _(b'writing %d bytes for %d files\n') % (bytecount, filecount)
        )

        yield struct.pack(b'>QQ', filecount, bytecount)
        yield struct.pack(b'>H', len(requires) + 1)
        yield requires + b'\0'

        # This is where we'll add compression in the future.
        assert compression == b'UN'

        progress = repo.ui.makeprogress(
            _(b'bundle'), total=bytecount, unit=_(b'bytes')
        )
        progress.update(0)

        for chunk in it:
            progress.increment(step=len(chunk))
            yield chunk

        progress.complete()

    return requirements, gen()


def consumev1(repo, fp, filecount, bytecount):
    """Apply the contents from version 1 of a streaming clone file handle.

    This takes the output from "stream_out" and applies it to the specified
    repository.

    Like "stream_out," the status line added by the wire protocol is not
    handled by this function.
    """
    with repo.lock():
        repo.ui.status(
            _(b'%d files to transfer, %s of data\n')
            % (filecount, util.bytecount(bytecount))
        )
        progress = repo.ui.makeprogress(
            _(b'clone'), total=bytecount, unit=_(b'bytes')
        )
        progress.update(0)
        start = util.timer()

        # TODO: get rid of (potential) inconsistency
        #
        # If transaction is started and any @filecache property is
        # changed at this point, it causes inconsistency between
        # in-memory cached property and streamclone-ed file on the
        # disk. Nested transaction prevents transaction scope "clone"
        # below from writing in-memory changes out at the end of it,
        # even though in-memory changes are discarded at the end of it
        # regardless of transaction nesting.
        #
        # But transaction nesting can't be simply prohibited, because
        # nesting occurs also in ordinary case (e.g. enabling
        # clonebundles).

        with repo.transaction(b'clone'):
            with repo.svfs.backgroundclosing(repo.ui, expectedcount=filecount):
                for i in pycompat.xrange(filecount):
                    # XXX doesn't support '\n' or '\r' in filenames
                    l = fp.readline()
                    try:
                        name, size = l.split(b'\0', 1)
                        size = int(size)
                    except (ValueError, TypeError):
                        raise error.ResponseError(
                            _(b'unexpected response from remote server:'), l
                        )
                    if repo.ui.debugflag:
                        repo.ui.debug(
                            b'adding %s (%s)\n' % (name, util.bytecount(size))
                        )
                    # for backwards compat, name was partially encoded
                    path = store.decodedir(name)
                    with repo.svfs(path, b'w', backgroundclose=True) as ofp:
                        for chunk in util.filechunkiter(fp, limit=size):
                            progress.increment(step=len(chunk))
                            ofp.write(chunk)

            # force @filecache properties to be reloaded from
            # streamclone-ed file at next access
            repo.invalidate(clearfilecache=True)

        elapsed = util.timer() - start
        if elapsed <= 0:
            elapsed = 0.001
        progress.complete()
        repo.ui.status(
            _(b'transferred %s in %.1f seconds (%s/sec)\n')
            % (
                util.bytecount(bytecount),
                elapsed,
                util.bytecount(bytecount / elapsed),
            )
        )


def readbundle1header(fp):
    compression = fp.read(2)
    if compression != b'UN':
        raise error.Abort(
            _(
                b'only uncompressed stream clone bundles are '
                b'supported; got %s'
            )
            % compression
        )

    filecount, bytecount = struct.unpack(b'>QQ', fp.read(16))
    requireslen = struct.unpack(b'>H', fp.read(2))[0]
    requires = fp.read(requireslen)

    if not requires.endswith(b'\0'):
        raise error.Abort(
            _(
                b'malformed stream clone bundle: '
                b'requirements not properly encoded'
            )
        )

    requirements = set(requires.rstrip(b'\0').split(b','))

    return filecount, bytecount, requirements


def applybundlev1(repo, fp):
    """Apply the content from a stream clone bundle version 1.

    We assume the 4 byte header has been read and validated and the file handle
    is at the 2 byte compression identifier.
    """
    if len(repo):
        raise error.Abort(
            _(b'cannot apply stream clone bundle on non-empty repo')
        )

    filecount, bytecount, requirements = readbundle1header(fp)
    missingreqs = requirements - repo.supportedformats
    if missingreqs:
        raise error.Abort(
            _(b'unable to apply stream clone: unsupported format: %s')
            % b', '.join(sorted(missingreqs))
        )

    consumev1(repo, fp, filecount, bytecount)


class streamcloneapplier(object):
    """Class to manage applying streaming clone bundles.

    We need to wrap ``applybundlev1()`` in a dedicated type to enable bundle
    readers to perform bundle type-specific functionality.
    """

    def __init__(self, fh):
        self._fh = fh

    def apply(self, repo):
        return applybundlev1(repo, self._fh)


# type of file to stream
_fileappend = 0  # append only file
_filefull = 1  # full snapshot file

# Source of the file
_srcstore = b's'  # store (svfs)
_srccache = b'c'  # cache (cache)

# This is it's own function so extensions can override it.
def _walkstreamfullstorefiles(repo):
    """list snapshot file from the store"""
    fnames = []
    if not repo.publishing():
        fnames.append(b'phaseroots')
    return fnames


def _filterfull(entry, copy, vfsmap):
    """actually copy the snapshot files"""
    src, name, ftype, data = entry
    if ftype != _filefull:
        return entry
    return (src, name, ftype, copy(vfsmap[src].join(name)))


@contextlib.contextmanager
def maketempcopies():
    """return a function to temporary copy file"""
    files = []
    try:

        def copy(src):
            fd, dst = pycompat.mkstemp()
            os.close(fd)
            files.append(dst)
            util.copyfiles(src, dst, hardlink=True)
            return dst

        yield copy
    finally:
        for tmp in files:
            util.tryunlink(tmp)


def _makemap(repo):
    """make a (src -> vfs) map for the repo"""
    vfsmap = {
        _srcstore: repo.svfs,
        _srccache: repo.cachevfs,
    }
    # we keep repo.vfs out of the on purpose, ther are too many danger there
    # (eg: .hg/hgrc)
    assert repo.vfs not in vfsmap.values()

    return vfsmap


def _emit2(repo, entries, totalfilesize):
    """actually emit the stream bundle"""
    vfsmap = _makemap(repo)
    # we keep repo.vfs out of the on purpose, ther are too many danger there
    # (eg: .hg/hgrc),
    #
    # this assert is duplicated (from _makemap) as author might think this is
    # fine, while this is really not fine.
    if repo.vfs in vfsmap.values():
        raise error.ProgrammingError(
            b'repo.vfs must not be added to vfsmap for security reasons'
        )

    progress = repo.ui.makeprogress(
        _(b'bundle'), total=totalfilesize, unit=_(b'bytes')
    )
    progress.update(0)
    with maketempcopies() as copy, progress:
        # copy is delayed until we are in the try
        entries = [_filterfull(e, copy, vfsmap) for e in entries]
        yield None  # this release the lock on the repository
        totalbytecount = 0

        for src, name, ftype, data in entries:
            vfs = vfsmap[src]
            yield src
            yield util.uvarintencode(len(name))
            if ftype == _fileappend:
                fp = vfs(name)
                size = data
            elif ftype == _filefull:
                fp = open(data, b'rb')
                size = util.fstat(fp).st_size
            bytecount = 0
            try:
                yield util.uvarintencode(size)
                yield name
                if size <= 65536:
                    chunks = (fp.read(size),)
                else:
                    chunks = util.filechunkiter(fp, limit=size)
                for chunk in chunks:
                    bytecount += len(chunk)
                    totalbytecount += len(chunk)
                    progress.update(totalbytecount)
                    yield chunk
                if bytecount != size:
                    # Would most likely be caused by a race due to `hg strip` or
                    # a revlog split
                    raise error.Abort(
                        _(
                            b'clone could only read %d bytes from %s, but '
                            b'expected %d bytes'
                        )
                        % (bytecount, name, size)
                    )
            finally:
                fp.close()


def _test_sync_point_walk_1(repo):
    """a function for synchronisation during tests"""


def _test_sync_point_walk_2(repo):
    """a function for synchronisation during tests"""


def _v2_walk(repo, includes, excludes, includeobsmarkers):
    """emit a seris of files information useful to clone a repo

    return (entries, totalfilesize)

    entries is a list of tuple (vfs-key, file-path, file-type, size)

    - `vfs-key`: is a key to the right vfs to write the file (see _makemap)
    - `name`: file path of the file to copy (to be feed to the vfss)
    - `file-type`: do this file need to be copied with the source lock ?
    - `size`: the size of the file (or None)
    """
    assert repo._currentlock(repo._lockref) is not None
    entries = []
    totalfilesize = 0

    matcher = None
    if includes or excludes:
        matcher = narrowspec.match(repo.root, includes, excludes)

    for rl_type, name, ename, size in _walkstreamfiles(repo, matcher):
        if size:
            ft = _fileappend
            if rl_type & store.FILEFLAGS_VOLATILE:
                ft = _filefull
            entries.append((_srcstore, name, ft, size))
            totalfilesize += size
    for name in _walkstreamfullstorefiles(repo):
        if repo.svfs.exists(name):
            totalfilesize += repo.svfs.lstat(name).st_size
            entries.append((_srcstore, name, _filefull, None))
    if includeobsmarkers and repo.svfs.exists(b'obsstore'):
        totalfilesize += repo.svfs.lstat(b'obsstore').st_size
        entries.append((_srcstore, b'obsstore', _filefull, None))
    for name in cacheutil.cachetocopy(repo):
        if repo.cachevfs.exists(name):
            totalfilesize += repo.cachevfs.lstat(name).st_size
            entries.append((_srccache, name, _filefull, None))
    return entries, totalfilesize


def generatev2(repo, includes, excludes, includeobsmarkers):
    """Emit content for version 2 of a streaming clone.

    the data stream consists the following entries:
    1) A char representing the file destination (eg: store or cache)
    2) A varint containing the length of the filename
    3) A varint containing the length of file data
    4) N bytes containing the filename (the internal, store-agnostic form)
    5) N bytes containing the file data

    Returns a 3-tuple of (file count, file size, data iterator).
    """

    with repo.lock():

        repo.ui.debug(b'scanning\n')

        entries, totalfilesize = _v2_walk(
            repo,
            includes=includes,
            excludes=excludes,
            includeobsmarkers=includeobsmarkers,
        )

        chunks = _emit2(repo, entries, totalfilesize)
        first = next(chunks)
        assert first is None
        _test_sync_point_walk_1(repo)
    _test_sync_point_walk_2(repo)

    return len(entries), totalfilesize, chunks


@contextlib.contextmanager
def nested(*ctxs):
    this = ctxs[0]
    rest = ctxs[1:]
    with this:
        if rest:
            with nested(*rest):
                yield
        else:
            yield


def consumev2(repo, fp, filecount, filesize):
    """Apply the contents from a version 2 streaming clone.

    Data is read from an object that only needs to provide a ``read(size)``
    method.
    """
    with repo.lock():
        repo.ui.status(
            _(b'%d files to transfer, %s of data\n')
            % (filecount, util.bytecount(filesize))
        )

        start = util.timer()
        progress = repo.ui.makeprogress(
            _(b'clone'), total=filesize, unit=_(b'bytes')
        )
        progress.update(0)

        vfsmap = _makemap(repo)
        # we keep repo.vfs out of the on purpose, ther are too many danger
        # there (eg: .hg/hgrc),
        #
        # this assert is duplicated (from _makemap) as author might think this
        # is fine, while this is really not fine.
        if repo.vfs in vfsmap.values():
            raise error.ProgrammingError(
                b'repo.vfs must not be added to vfsmap for security reasons'
            )

        with repo.transaction(b'clone'):
            ctxs = (vfs.backgroundclosing(repo.ui) for vfs in vfsmap.values())
            with nested(*ctxs):
                for i in range(filecount):
                    src = util.readexactly(fp, 1)
                    vfs = vfsmap[src]
                    namelen = util.uvarintdecodestream(fp)
                    datalen = util.uvarintdecodestream(fp)

                    name = util.readexactly(fp, namelen)

                    if repo.ui.debugflag:
                        repo.ui.debug(
                            b'adding [%s] %s (%s)\n'
                            % (src, name, util.bytecount(datalen))
                        )

                    with vfs(name, b'w') as ofp:
                        for chunk in util.filechunkiter(fp, limit=datalen):
                            progress.increment(step=len(chunk))
                            ofp.write(chunk)

            # force @filecache properties to be reloaded from
            # streamclone-ed file at next access
            repo.invalidate(clearfilecache=True)

        elapsed = util.timer() - start
        if elapsed <= 0:
            elapsed = 0.001
        repo.ui.status(
            _(b'transferred %s in %.1f seconds (%s/sec)\n')
            % (
                util.bytecount(progress.pos),
                elapsed,
                util.bytecount(progress.pos / elapsed),
            )
        )
        progress.complete()


def applybundlev2(repo, fp, filecount, filesize, requirements):
    from . import localrepo

    missingreqs = [r for r in requirements if r not in repo.supported]
    if missingreqs:
        raise error.Abort(
            _(b'unable to apply stream clone: unsupported format: %s')
            % b', '.join(sorted(missingreqs))
        )

    consumev2(repo, fp, filecount, filesize)

    # new requirements = old non-format requirements +
    #                    new format-related remote requirements
    # requirements from the streamed-in repository
    repo.requirements = set(requirements) | (
        repo.requirements - repo.supportedformats
    )
    repo.svfs.options = localrepo.resolvestorevfsoptions(
        repo.ui, repo.requirements, repo.features
    )
    scmutil.writereporequirements(repo)


def _copy_files(src_vfs_map, dst_vfs_map, entries, progress):
    hardlink = [True]

    def copy_used():
        hardlink[0] = False
        progress.topic = _(b'copying')

    for k, path, size in entries:
        src_vfs = src_vfs_map[k]
        dst_vfs = dst_vfs_map[k]
        src_path = src_vfs.join(path)
        dst_path = dst_vfs.join(path)
        # We cannot use dirname and makedirs of dst_vfs here because the store
        # encoding confuses them. See issue 6581 for details.
        dirname = os.path.dirname(dst_path)
        if not os.path.exists(dirname):
            util.makedirs(dirname)
        dst_vfs.register_file(path)
        # XXX we could use the #nb_bytes argument.
        util.copyfile(
            src_path,
            dst_path,
            hardlink=hardlink[0],
            no_hardlink_cb=copy_used,
            check_fs_hardlink=False,
        )
        progress.increment()
    return hardlink[0]


def local_copy(src_repo, dest_repo):
    """copy all content from one local repository to another

    This is useful for local clone"""
    src_store_requirements = {
        r
        for r in src_repo.requirements
        if r not in requirementsmod.WORKING_DIR_REQUIREMENTS
    }
    dest_store_requirements = {
        r
        for r in dest_repo.requirements
        if r not in requirementsmod.WORKING_DIR_REQUIREMENTS
    }
    assert src_store_requirements == dest_store_requirements

    with dest_repo.lock():
        with src_repo.lock():

            # bookmark is not integrated to the streaming as it might use the
            # `repo.vfs` and they are too many sentitive data accessible
            # through `repo.vfs` to expose it to streaming clone.
            src_book_vfs = bookmarks.bookmarksvfs(src_repo)
            srcbookmarks = src_book_vfs.join(b'bookmarks')
            bm_count = 0
            if os.path.exists(srcbookmarks):
                bm_count = 1

            entries, totalfilesize = _v2_walk(
                src_repo,
                includes=None,
                excludes=None,
                includeobsmarkers=True,
            )
            src_vfs_map = _makemap(src_repo)
            dest_vfs_map = _makemap(dest_repo)
            progress = src_repo.ui.makeprogress(
                topic=_(b'linking'),
                total=len(entries) + bm_count,
                unit=_(b'files'),
            )
            # copy  files
            #
            # We could copy the full file while the source repository is locked
            # and the other one without the lock. However, in the linking case,
            # this would also requires checks that nobody is appending any data
            # to the files while we do the clone, so this is not done yet. We
            # could do this blindly when copying files.
            files = ((k, path, size) for k, path, ftype, size in entries)
            hardlink = _copy_files(src_vfs_map, dest_vfs_map, files, progress)

            # copy bookmarks over
            if bm_count:
                dst_book_vfs = bookmarks.bookmarksvfs(dest_repo)
                dstbookmarks = dst_book_vfs.join(b'bookmarks')
                util.copyfile(srcbookmarks, dstbookmarks)
        progress.complete()
        if hardlink:
            msg = b'linked %d files\n'
        else:
            msg = b'copied %d files\n'
        src_repo.ui.debug(msg % (len(entries) + bm_count))

        with dest_repo.transaction(b"localclone") as tr:
            dest_repo.store.write(tr)

        # clean up transaction file as they do not make sense
        undo_files = [(dest_repo.svfs, b'undo.backupfiles')]
        undo_files.extend(dest_repo.undofiles())
        for undovfs, undofile in undo_files:
            try:
                undovfs.unlink(undofile)
            except OSError as e:
                if e.errno != errno.ENOENT:
                    msg = _(b'error removing %s: %s\n')
                    path = undovfs.join(undofile)
                    e_msg = stringutil.forcebytestr(e)
                    msg %= (path, e_msg)
                    dest_repo.ui.warn(msg)



# manifest.py - manifest revision class for mercurial
#
# Copyright 2005-2007 Olivia Mackall <olivia@selenic.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import absolute_import

import heapq
import itertools
import struct
import weakref

from .i18n import _
from .node import (
    bin,
    hex,
    nullrev,
)
from .pycompat import getattr
from . import (
    encoding,
    error,
    match as matchmod,
    mdiff,
    pathutil,
    policy,
    pycompat,
    revlog,
    util,
)
from .interfaces import (
    repository,
    util as interfaceutil,
)
from .revlogutils import (
    constants as revlog_constants,
)

parsers = policy.importmod('parsers')
propertycache = util.propertycache

# Allow tests to more easily test the alternate path in manifestdict.fastdelta()
FASTDELTA_TEXTDIFF_THRESHOLD = 1000


def _parse(nodelen, data):
    # This method does a little bit of excessive-looking
    # precondition checking. This is so that the behavior of this
    # class exactly matches its C counterpart to try and help
    # prevent surprise breakage for anyone that develops against
    # the pure version.
    if data and data[-1:] != b'\n':
        raise ValueError(b'Manifest did not end in a newline.')
    prev = None
    for l in data.splitlines():
        if prev is not None and prev > l:
            raise ValueError(b'Manifest lines not in sorted order.')
        prev = l
        f, n = l.split(b'\0')
        nl = len(n)
        flags = n[-1:]
        if flags in _manifestflags:
            n = n[:-1]
            nl -= 1
        else:
            flags = b''
        if nl != 2 * nodelen:
            raise ValueError(b'Invalid manifest line')

        yield f, bin(n), flags


def _text(it):
    files = []
    lines = []
    for f, n, fl in it:
        files.append(f)
        # if this is changed to support newlines in filenames,
        # be sure to check the templates/ dir again (especially *-raw.tmpl)
        lines.append(b"%s\0%s%s\n" % (f, hex(n), fl))

    _checkforbidden(files)
    return b''.join(lines)


class lazymanifestiter(object):
    def __init__(self, lm):
        self.pos = 0
        self.lm = lm

    def __iter__(self):
        return self

    def next(self):
        try:
            data, pos = self.lm._get(self.pos)
        except IndexError:
            raise StopIteration
        if pos == -1:
            self.pos += 1
            return data[0]
        self.pos += 1
        zeropos = data.find(b'\x00', pos)
        return data[pos:zeropos]

    __next__ = next


class lazymanifestiterentries(object):
    def __init__(self, lm):
        self.lm = lm
        self.pos = 0

    def __iter__(self):
        return self

    def next(self):
        try:
            data, pos = self.lm._get(self.pos)
        except IndexError:
            raise StopIteration
        if pos == -1:
            self.pos += 1
            return data
        zeropos = data.find(b'\x00', pos)
        nlpos = data.find(b'\n', pos)
        if zeropos == -1 or nlpos == -1 or nlpos < zeropos:
            raise error.StorageError(b'Invalid manifest line')
        flags = data[nlpos - 1 : nlpos]
        if flags in _manifestflags:
            hlen = nlpos - zeropos - 2
        else:
            hlen = nlpos - zeropos - 1
            flags = b''
        if hlen != 2 * self.lm._nodelen:
            raise error.StorageError(b'Invalid manifest line')
        hashval = unhexlify(
            data, self.lm.extrainfo[self.pos], zeropos + 1, hlen
        )
        self.pos += 1
        return (data[pos:zeropos], hashval, flags)

    __next__ = next


def unhexlify(data, extra, pos, length):
    s = bin(data[pos : pos + length])
    if extra:
        s += chr(extra & 0xFF)
    return s


def _cmp(a, b):
    return (a > b) - (a < b)


_manifestflags = {b'', b'l', b't', b'x'}


class _lazymanifest(object):
    """A pure python manifest backed by a byte string.  It is supplimented with
    internal lists as it is modified, until it is compacted back to a pure byte
    string.

    ``data`` is the initial manifest data.

    ``positions`` is a list of offsets, one per manifest entry.  Positive
    values are offsets into ``data``, negative values are offsets into the
    ``extradata`` list.  When an entry is removed, its entry is dropped from
    ``positions``.  The values are encoded such that when walking the list and
    indexing into ``data`` or ``extradata`` as appropriate, the entries are
    sorted by filename.

    ``extradata`` is a list of (key, hash, flags) for entries that were added or
    modified since the manifest was created or compacted.
    """

    def __init__(
        self,
        nodelen,
        data,
        positions=None,
        extrainfo=None,
        extradata=None,
        hasremovals=False,
    ):
        self._nodelen = nodelen
        if positions is None:
            self.positions = self.findlines(data)
            self.extrainfo = [0] * len(self.positions)
            self.data = data
            self.extradata = []
            self.hasremovals = False
        else:
            self.positions = positions[:]
            self.extrainfo = extrainfo[:]
            self.extradata = extradata[:]
            self.data = data
            self.hasremovals = hasremovals

    def findlines(self, data):
        if not data:
            return []
        pos = data.find(b"\n")
        if pos == -1 or data[-1:] != b'\n':
            raise ValueError(b"Manifest did not end in a newline.")
        positions = [0]
        prev = data[: data.find(b'\x00')]
        while pos < len(data) - 1 and pos != -1:
            positions.append(pos + 1)
            nexts = data[pos + 1 : data.find(b'\x00', pos + 1)]
            if nexts < prev:
                raise ValueError(b"Manifest lines not in sorted order.")
            prev = nexts
            pos = data.find(b"\n", pos + 1)
        return positions

    def _get(self, index):
        # get the position encoded in pos:
        #   positive number is an index in 'data'
        #   negative number is in extrapieces
        pos = self.positions[index]
        if pos >= 0:
            return self.data, pos
        return self.extradata[-pos - 1], -1

    def _getkey(self, pos):
        if pos >= 0:
            return self.data[pos : self.data.find(b'\x00', pos + 1)]
        return self.extradata[-pos - 1][0]

    def bsearch(self, key):
        first = 0
        last = len(self.positions) - 1

        while first <= last:
            midpoint = (first + last) // 2
            nextpos = self.positions[midpoint]
            candidate = self._getkey(nextpos)
            r = _cmp(key, candidate)
            if r == 0:
                return midpoint
            else:
                if r < 0:
                    last = midpoint - 1
                else:
                    first = midpoint + 1
        return -1

    def bsearch2(self, key):
        # same as the above, but will always return the position
        # done for performance reasons
        first = 0
        last = len(self.positions) - 1

        while first <= last:
            midpoint = (first + last) // 2
            nextpos = self.positions[midpoint]
            candidate = self._getkey(nextpos)
            r = _cmp(key, candidate)
            if r == 0:
                return (midpoint, True)
            else:
                if r < 0:
                    last = midpoint - 1
                else:
                    first = midpoint + 1
        return (first, False)

    def __contains__(self, key):
        return self.bsearch(key) != -1

    def __getitem__(self, key):
        if not isinstance(key, bytes):
            raise TypeError(b"getitem: manifest keys must be a bytes.")
        needle = self.bsearch(key)
        if needle == -1:
            raise KeyError
        data, pos = self._get(needle)
        if pos == -1:
            return (data[1], data[2])
        zeropos = data.find(b'\x00', pos)
        nlpos = data.find(b'\n', zeropos)
        assert 0 <= needle <= len(self.positions)
        assert len(self.extrainfo) == len(self.positions)
        if zeropos == -1 or nlpos == -1 or nlpos < zeropos:
            raise error.StorageError(b'Invalid manifest line')
        hlen = nlpos - zeropos - 1
        flags = data[nlpos - 1 : nlpos]
        if flags in _manifestflags:
            hlen -= 1
        else:
            flags = b''
        if hlen != 2 * self._nodelen:
            raise error.StorageError(b'Invalid manifest line')
        hashval = unhexlify(data, self.extrainfo[needle], zeropos + 1, hlen)
        return (hashval, flags)

    def __delitem__(self, key):
        needle, found = self.bsearch2(key)
        if not found:
            raise KeyError
        cur = self.positions[needle]
        self.positions = self.positions[:needle] + self.positions[needle + 1 :]
        self.extrainfo = self.extrainfo[:needle] + self.extrainfo[needle + 1 :]
        if cur >= 0:
            # This does NOT unsort the list as far as the search functions are
            # concerned, as they only examine lines mapped by self.positions.
            self.data = self.data[:cur] + b'\x00' + self.data[cur + 1 :]
            self.hasremovals = True

    def __setitem__(self, key, value):
        if not isinstance(key, bytes):
            raise TypeError(b"setitem: manifest keys must be a byte string.")
        if not isinstance(value, tuple) or len(value) != 2:
            raise TypeError(
                b"Manifest values must be a tuple of (node, flags)."
            )
        hashval = value[0]
        if not isinstance(hashval, bytes) or len(hashval) not in (20, 32):
            raise TypeError(b"node must be a 20-byte or 32-byte byte string")
        flags = value[1]
        if not isinstance(flags, bytes) or len(flags) > 1:
            raise TypeError(b"flags must a 0 or 1 byte string, got %r", flags)
        needle, found = self.bsearch2(key)
        if found:
            # put the item
            pos = self.positions[needle]
            if pos < 0:
                self.extradata[-pos - 1] = (key, hashval, value[1])
            else:
                # just don't bother
                self.extradata.append((key, hashval, value[1]))
                self.positions[needle] = -len(self.extradata)
        else:
            # not found, put it in with extra positions
            self.extradata.append((key, hashval, value[1]))
            self.positions = (
                self.positions[:needle]
                + [-len(self.extradata)]
                + self.positions[needle:]
            )
            self.extrainfo = (
                self.extrainfo[:needle] + [0] + self.extrainfo[needle:]
            )

    def copy(self):
        # XXX call _compact like in C?
        return _lazymanifest(
            self._nodelen,
            self.data,
            self.positions,
            self.extrainfo,
            self.extradata,
            self.hasremovals,
        )

    def _compact(self):
        # hopefully not called TOO often
        if len(self.extradata) == 0 and not self.hasremovals:
            return
        l = []
        i = 0
        offset = 0
        self.extrainfo = [0] * len(self.positions)
        while i < len(self.positions):
            if self.positions[i] >= 0:
                cur = self.positions[i]
                last_cut = cur

                # Collect all contiguous entries in the buffer at the current
                # offset, breaking out only for added/modified items held in
                # extradata, or a deleted line prior to the next position.
                while True:
                    self.positions[i] = offset
                    i += 1
                    if i == len(self.positions) or self.positions[i] < 0:
                        break

                    # A removed file has no positions[] entry, but does have an
                    # overwritten first byte.  Break out and find the end of the
                    # current good entry/entries if there is a removed file
                    # before the next position.
                    if (
                        self.hasremovals
                        and self.data.find(b'\n\x00', cur, self.positions[i])
                        != -1
                    ):
                        break

                    offset += self.positions[i] - cur
                    cur = self.positions[i]
                end_cut = self.data.find(b'\n', cur)
                if end_cut != -1:
                    end_cut += 1
                offset += end_cut - cur
                l.append(self.data[last_cut:end_cut])
            else:
                while i < len(self.positions) and self.positions[i] < 0:
                    cur = self.positions[i]
                    t = self.extradata[-cur - 1]
                    l.append(self._pack(t))
                    self.positions[i] = offset
                    # Hashes are either 20 bytes (old sha1s) or 32
                    # bytes (new non-sha1).
                    hlen = 20
                    if len(t[1]) > 25:
                        hlen = 32
                    if len(t[1]) > hlen:
                        self.extrainfo[i] = ord(t[1][hlen + 1])
                    offset += len(l[-1])
                    i += 1
        self.data = b''.join(l)
        self.hasremovals = False
        self.extradata = []

    def _pack(self, d):
        n = d[1]
        assert len(n) in (20, 32)
        return d[0] + b'\x00' + hex(n) + d[2] + b'\n'

    def text(self):
        self._compact()
        return self.data

    def diff(self, m2, clean=False):
        '''Finds changes between the current manifest and m2.'''
        # XXX think whether efficiency matters here
        diff = {}

        for fn, e1, flags in self.iterentries():
            if fn not in m2:
                diff[fn] = (e1, flags), (None, b'')
            else:
                e2 = m2[fn]
                if (e1, flags) != e2:
                    diff[fn] = (e1, flags), e2
                elif clean:
                    diff[fn] = None

        for fn, e2, flags in m2.iterentries():
            if fn not in self:
                diff[fn] = (None, b''), (e2, flags)

        return diff

    def iterentries(self):
        return lazymanifestiterentries(self)

    def iterkeys(self):
        return lazymanifestiter(self)

    def __iter__(self):
        return lazymanifestiter(self)

    def __len__(self):
        return len(self.positions)

    def filtercopy(self, filterfn):
        # XXX should be optimized
        c = _lazymanifest(self._nodelen, b'')
        for f, n, fl in self.iterentries():
            if filterfn(f):
                c[f] = n, fl
        return c


try:
    _lazymanifest = parsers.lazymanifest
except AttributeError:
    pass


@interfaceutil.implementer(repository.imanifestdict)
class manifestdict(object):
    def __init__(self, nodelen, data=b''):
        self._nodelen = nodelen
        self._lm = _lazymanifest(nodelen, data)

    def __getitem__(self, key):
        return self._lm[key][0]

    def find(self, key):
        return self._lm[key]

    def __len__(self):
        return len(self._lm)

    def __nonzero__(self):
        # nonzero is covered by the __len__ function, but implementing it here
        # makes it easier for extensions to override.
        return len(self._lm) != 0

    __bool__ = __nonzero__

    def __setitem__(self, key, node):
        self._lm[key] = node, self.flags(key)

    def __contains__(self, key):
        if key is None:
            return False
        return key in self._lm

    def __delitem__(self, key):
        del self._lm[key]

    def __iter__(self):
        return self._lm.__iter__()

    def iterkeys(self):
        return self._lm.iterkeys()

    def keys(self):
        return list(self.iterkeys())

    def filesnotin(self, m2, match=None):
        '''Set of files in this manifest that are not in the other'''
        if match is not None:
            match = matchmod.badmatch(match, lambda path, msg: None)
            sm2 = set(m2.walk(match))
            return {f for f in self.walk(match) if f not in sm2}
        return {f for f in self if f not in m2}

    @propertycache
    def _dirs(self):
        return pathutil.dirs(self)

    def dirs(self):
        return self._dirs

    def hasdir(self, dir):
        return dir in self._dirs

    def _filesfastpath(self, match):
        """Checks whether we can correctly and quickly iterate over matcher
        files instead of over manifest files."""
        files = match.files()
        return len(files) < 100 and (
            match.isexact()
            or (match.prefix() and all(fn in self for fn in files))
        )

    def walk(self, match):
        """Generates matching file names.

        Equivalent to manifest.matches(match).iterkeys(), but without creating
        an entirely new manifest.

        It also reports nonexistent files by marking them bad with match.bad().
        """
        if match.always():
            for f in iter(self):
                yield f
            return

        fset = set(match.files())

        # avoid the entire walk if we're only looking for specific files
        if self._filesfastpath(match):
            for fn in sorted(fset):
                if fn in self:
                    yield fn
            return

        for fn in self:
            if fn in fset:
                # specified pattern is the exact name
                fset.remove(fn)
            if match(fn):
                yield fn

        # for dirstate.walk, files=[''] means "walk the whole tree".
        # follow that here, too
        fset.discard(b'')

        for fn in sorted(fset):
            if not self.hasdir(fn):
                match.bad(fn, None)

    def _matches(self, match):
        '''generate a new manifest filtered by the match argument'''
        if match.always():
            return self.copy()

        if self._filesfastpath(match):
            m = manifestdict(self._nodelen)
            lm = self._lm
            for fn in match.files():
                if fn in lm:
                    m._lm[fn] = lm[fn]
            return m

        m = manifestdict(self._nodelen)
        m._lm = self._lm.filtercopy(match)
        return m

    def diff(self, m2, match=None, clean=False):
        """Finds changes between the current manifest and m2.

        Args:
          m2: the manifest to which this manifest should be compared.
          clean: if true, include files unchanged between these manifests
                 with a None value in the returned dictionary.

        The result is returned as a dict with filename as key and
        values of the form ((n1,fl1),(n2,fl2)), where n1/n2 is the
        nodeid in the current/other manifest and fl1/fl2 is the flag
        in the current/other manifest. Where the file does not exist,
        the nodeid will be None and the flags will be the empty
        string.
        """
        if match:
            m1 = self._matches(match)
            m2 = m2._matches(match)
            return m1.diff(m2, clean=clean)
        return self._lm.diff(m2._lm, clean)

    def setflag(self, key, flag):
        if flag not in _manifestflags:
            raise TypeError(b"Invalid manifest flag set.")
        self._lm[key] = self[key], flag

    def get(self, key, default=None):
        try:
            return self._lm[key][0]
        except KeyError:
            return default

    def flags(self, key):
        try:
            return self._lm[key][1]
        except KeyError:
            return b''

    def copy(self):
        c = manifestdict(self._nodelen)
        c._lm = self._lm.copy()
        return c

    def items(self):
        return (x[:2] for x in self._lm.iterentries())

    def iteritems(self):
        return (x[:2] for x in self._lm.iterentries())

    def iterentries(self):
        return self._lm.iterentries()

    def text(self):
        # most likely uses native version
        return self._lm.text()

    def fastdelta(self, base, changes):
        """Given a base manifest text as a bytearray and a list of changes
        relative to that text, compute a delta that can be used by revlog.
        """
        delta = []
        dstart = None
        dend = None
        dline = [b""]
        start = 0
        # zero copy representation of base as a buffer
        addbuf = util.buffer(base)

        changes = list(changes)
        if len(changes) < FASTDELTA_TEXTDIFF_THRESHOLD:
            # start with a readonly loop that finds the offset of
            # each line and creates the deltas
            for f, todelete in changes:
                # bs will either be the index of the item or the insert point
                start, end = _msearch(addbuf, f, start)
                if not todelete:
                    h, fl = self._lm[f]
                    l = b"%s\0%s%s\n" % (f, hex(h), fl)
                else:
                    if start == end:
                        # item we want to delete was not found, error out
                        raise AssertionError(
                            _(b"failed to remove %s from manifest") % f
                        )
                    l = b""
                if dstart is not None and dstart <= start and dend >= start:
                    if dend < end:
                        dend = end
                    if l:
                        dline.append(l)
                else:
                    if dstart is not None:
                        delta.append([dstart, dend, b"".join(dline)])
                    dstart = start
                    dend = end
                    dline = [l]

            if dstart is not None:
                delta.append([dstart, dend, b"".join(dline)])
            # apply the delta to the base, and get a delta for addrevision
            deltatext, arraytext = _addlistdelta(base, delta)
        else:
            # For large changes, it's much cheaper to just build the text and
            # diff it.
            arraytext = bytearray(self.text())
            deltatext = mdiff.textdiff(
                util.buffer(base), util.buffer(arraytext)
            )

        return arraytext, deltatext


def _msearch(m, s, lo=0, hi=None):
    """return a tuple (start, end) that says where to find s within m.

    If the string is found m[start:end] are the line containing
    that string.  If start == end the string was not found and
    they indicate the proper sorted insertion point.

    m should be a buffer, a memoryview or a byte string.
    s is a byte string"""

    def advance(i, c):
        while i < lenm and m[i : i + 1] != c:
            i += 1
        return i

    if not s:
        return (lo, lo)
    lenm = len(m)
    if not hi:
        hi = lenm
    while lo < hi:
        mid = (lo + hi) // 2
        start = mid
        while start > 0 and m[start - 1 : start] != b'\n':
            start -= 1
        end = advance(start, b'\0')
        if bytes(m[start:end]) < s:
            # we know that after the null there are 40 bytes of sha1
            # this translates to the bisect lo = mid + 1
            lo = advance(end + 40, b'\n') + 1
        else:
            # this translates to the bisect hi = mid
            hi = start
    end = advance(lo, b'\0')
    found = m[lo:end]
    if s == found:
        # we know that after the null there are 40 bytes of sha1
        end = advance(end + 40, b'\n')
        return (lo, end + 1)
    else:
        return (lo, lo)


def _checkforbidden(l):
    """Check filenames for illegal characters."""
    for f in l:
        if b'\n' in f or b'\r' in f:
            raise error.StorageError(
                _(b"'\\n' and '\\r' disallowed in filenames: %r")
                % pycompat.bytestr(f)
            )


# apply the changes collected during the bisect loop to our addlist
# return a delta suitable for addrevision
def _addlistdelta(addlist, x):
    # for large addlist arrays, building a new array is cheaper
    # than repeatedly modifying the existing one
    currentposition = 0
    newaddlist = bytearray()

    for start, end, content in x:
        newaddlist += addlist[currentposition:start]
        if content:
            newaddlist += bytearray(content)

        currentposition = end

    newaddlist += addlist[currentposition:]

    deltatext = b"".join(
        struct.pack(b">lll", start, end, len(content)) + content
        for start, end, content in x
    )
    return deltatext, newaddlist


def _splittopdir(f):
    if b'/' in f:
        dir, subpath = f.split(b'/', 1)
        return dir + b'/', subpath
    else:
        return b'', f


_noop = lambda s: None


@interfaceutil.implementer(repository.imanifestdict)
class treemanifest(object):
    def __init__(self, nodeconstants, dir=b'', text=b''):
        self._dir = dir
        self.nodeconstants = nodeconstants
        self._node = self.nodeconstants.nullid
        self._nodelen = self.nodeconstants.nodelen
        self._loadfunc = _noop
        self._copyfunc = _noop
        self._dirty = False
        self._dirs = {}
        self._lazydirs = {}
        # Using _lazymanifest here is a little slower than plain old dicts
        self._files = {}
        self._flags = {}
        if text:

            def readsubtree(subdir, subm):
                raise AssertionError(
                    b'treemanifest constructor only accepts flat manifests'
                )

            self.parse(text, readsubtree)
            self._dirty = True  # Mark flat manifest dirty after parsing

    def _subpath(self, path):
        return self._dir + path

    def _loadalllazy(self):
        selfdirs = self._dirs
        subpath = self._subpath
        for d, (node, readsubtree, docopy) in pycompat.iteritems(
            self._lazydirs
        ):
            if docopy:
                selfdirs[d] = readsubtree(subpath(d), node).copy()
            else:
                selfdirs[d] = readsubtree(subpath(d), node)
        self._lazydirs = {}

    def _loadlazy(self, d):
        v = self._lazydirs.get(d)
        if v:
            node, readsubtree, docopy = v
            if docopy:
                self._dirs[d] = readsubtree(self._subpath(d), node).copy()
            else:
                self._dirs[d] = readsubtree(self._subpath(d), node)
            del self._lazydirs[d]

    def _loadchildrensetlazy(self, visit):
        if not visit:
            return None
        if visit == b'all' or visit == b'this':
            self._loadalllazy()
            return None

        loadlazy = self._loadlazy
        for k in visit:
            loadlazy(k + b'/')
        return visit

    def _loaddifflazy(self, t1, t2):
        """load items in t1 and t2 if they're needed for diffing.

        The criteria currently is:
        - if it's not present in _lazydirs in either t1 or t2, load it in the
          other (it may already be loaded or it may not exist, doesn't matter)
        - if it's present in _lazydirs in both, compare the nodeid; if it
          differs, load it in both
        """
        toloadlazy = []
        for d, v1 in pycompat.iteritems(t1._lazydirs):
            v2 = t2._lazydirs.get(d)
            if not v2 or v2[0] != v1[0]:
                toloadlazy.append(d)
        for d, v1 in pycompat.iteritems(t2._lazydirs):
            if d not in t1._lazydirs:
                toloadlazy.append(d)

        for d in toloadlazy:
            t1._loadlazy(d)
            t2._loadlazy(d)

    def __len__(self):
        self._load()
        size = len(self._files)
        self._loadalllazy()
        for m in self._dirs.values():
            size += m.__len__()
        return size

    def __nonzero__(self):
        # Faster than "__len() != 0" since it avoids loading sub-manifests
        return not self._isempty()

    __bool__ = __nonzero__

    def _isempty(self):
        self._load()  # for consistency; already loaded by all callers
        # See if we can skip loading everything.
        if self._files or (
            self._dirs and any(not m._isempty() for m in self._dirs.values())
        ):
            return False
        self._loadalllazy()
        return not self._dirs or all(m._isempty() for m in self._dirs.values())

    @encoding.strmethod
    def __repr__(self):
        return (
            b'<treemanifest dir=%s, node=%s, loaded=%r, dirty=%r at 0x%x>'
            % (
                self._dir,
                hex(self._node),
                bool(self._loadfunc is _noop),
                self._dirty,
                id(self),
            )
        )

    def dir(self):
        """The directory that this tree manifest represents, including a
        trailing '/'. Empty string for the repo root directory."""
        return self._dir

    def node(self):
        """This node of this instance. nullid for unsaved instances. Should
        be updated when the instance is read or written from a revlog.
        """
        assert not self._dirty
        return self._node

    def setnode(self, node):
        self._node = node
        self._dirty = False

    def iterentries(self):
        self._load()
        self._loadalllazy()
        for p, n in sorted(
            itertools.chain(self._dirs.items(), self._files.items())
        ):
            if p in self._files:
                yield self._subpath(p), n, self._flags.get(p, b'')
            else:
                for x in n.iterentries():
                    yield x

    def items(self):
        self._load()
        self._loadalllazy()
        for p, n in sorted(
            itertools.chain(self._dirs.items(), self._files.items())
        ):
            if p in self._files:
                yield self._subpath(p), n
            else:
                for f, sn in pycompat.iteritems(n):
                    yield f, sn

    iteritems = items

    def iterkeys(self):
        self._load()
        self._loadalllazy()
        for p in sorted(itertools.chain(self._dirs, self._files)):
            if p in self._files:
                yield self._subpath(p)
            else:
                for f in self._dirs[p]:
                    yield f

    def keys(self):
        return list(self.iterkeys())

    def __iter__(self):
        return self.iterkeys()

    def __contains__(self, f):
        if f is None:
            return False
        self._load()
        dir, subpath = _splittopdir(f)
        if dir:
            self._loadlazy(dir)

            if dir not in self._dirs:
                return False

            return self._dirs[dir].__contains__(subpath)
        else:
            return f in self._files

    def get(self, f, default=None):
        self._load()
        dir, subpath = _splittopdir(f)
        if dir:
            self._loadlazy(dir)

            if dir not in self._dirs:
                return default
            return self._dirs[dir].get(subpath, default)
        else:
            return self._files.get(f, default)

    def __getitem__(self, f):
        self._load()
        dir, subpath = _splittopdir(f)
        if dir:
            self._loadlazy(dir)

            return self._dirs[dir].__getitem__(subpath)
        else:
            return self._files[f]

    def flags(self, f):
        self._load()
        dir, subpath = _splittopdir(f)
        if dir:
            self._loadlazy(dir)

            if dir not in self._dirs:
                return b''
            return self._dirs[dir].flags(subpath)
        else:
            if f in self._lazydirs or f in self._dirs:
                return b''
            return self._flags.get(f, b'')

    def find(self, f):
        self._load()
        dir, subpath = _splittopdir(f)
        if dir:
            self._loadlazy(dir)

            return self._dirs[dir].find(subpath)
        else:
            return self._files[f], self._flags.get(f, b'')

    def __delitem__(self, f):
        self._load()
        dir, subpath = _splittopdir(f)
        if dir:
            self._loadlazy(dir)

            self._dirs[dir].__delitem__(subpath)
            # If the directory is now empty, remove it
            if self._dirs[dir]._isempty():
                del self._dirs[dir]
        else:
            del self._files[f]
            if f in self._flags:
                del self._flags[f]
        self._dirty = True

    def __setitem__(self, f, n):
        assert n is not None
        self._load()
        dir, subpath = _splittopdir(f)
        if dir:
            self._loadlazy(dir)
            if dir not in self._dirs:
                self._dirs[dir] = treemanifest(
                    self.nodeconstants, self._subpath(dir)
                )
            self._dirs[dir].__setitem__(subpath, n)
        else:
            # manifest nodes are either 20 bytes or 32 bytes,
            # depending on the hash in use. Assert this as historically
            # sometimes extra bytes were added.
            assert len(n) in (20, 32)
            self._files[f] = n
        self._dirty = True

    def _load(self):
        if self._loadfunc is not _noop:
            lf, self._loadfunc = self._loadfunc, _noop
            lf(self)
        elif self._copyfunc is not _noop:
            cf, self._copyfunc = self._copyfunc, _noop
            cf(self)

    def setflag(self, f, flags):
        """Set the flags (symlink, executable) for path f."""
        if flags not in _manifestflags:
            raise TypeError(b"Invalid manifest flag set.")
        self._load()
        dir, subpath = _splittopdir(f)
        if dir:
            self._loadlazy(dir)
            if dir not in self._dirs:
                self._dirs[dir] = treemanifest(
                    self.nodeconstants, self._subpath(dir)
                )
            self._dirs[dir].setflag(subpath, flags)
        else:
            self._flags[f] = flags
        self._dirty = True

    def copy(self):
        copy = treemanifest(self.nodeconstants, self._dir)
        copy._node = self._node
        copy._dirty = self._dirty
        if self._copyfunc is _noop:

            def _copyfunc(s):
                self._load()
                s._lazydirs = {
                    d: (n, r, True)
                    for d, (n, r, c) in pycompat.iteritems(self._lazydirs)
                }
                sdirs = s._dirs
                for d, v in pycompat.iteritems(self._dirs):
                    sdirs[d] = v.copy()
                s._files = dict.copy(self._files)
                s._flags = dict.copy(self._flags)

            if self._loadfunc is _noop:
                _copyfunc(copy)
            else:
                copy._copyfunc = _copyfunc
        else:
            copy._copyfunc = self._copyfunc
        return copy

    def filesnotin(self, m2, match=None):
        '''Set of files in this manifest that are not in the other'''
        if match and not match.always():
            m1 = self._matches(match)
            m2 = m2._matches(match)
            return m1.filesnotin(m2)

        files = set()

        def _filesnotin(t1, t2):
            if t1._node == t2._node and not t1._dirty and not t2._dirty:
                return
            t1._load()
            t2._load()
            self._loaddifflazy(t1, t2)
            for d, m1 in pycompat.iteritems(t1._dirs):
                if d in t2._dirs:
                    m2 = t2._dirs[d]
                    _filesnotin(m1, m2)
                else:
                    files.update(m1.iterkeys())

            for fn in t1._files:
                if fn not in t2._files:
                    files.add(t1._subpath(fn))

        _filesnotin(self, m2)
        return files

    @propertycache
    def _alldirs(self):
        return pathutil.dirs(self)

    def dirs(self):
        return self._alldirs

    def hasdir(self, dir):
        self._load()
        topdir, subdir = _splittopdir(dir)
        if topdir:
            self._loadlazy(topdir)
            if topdir in self._dirs:
                return self._dirs[topdir].hasdir(subdir)
            return False
        dirslash = dir + b'/'
        return dirslash in self._dirs or dirslash in self._lazydirs

    def walk(self, match):
        """Generates matching file names.

        It also reports nonexistent files by marking them bad with match.bad().
        """
        if match.always():
            for f in iter(self):
                yield f
            return

        fset = set(match.files())

        for fn in self._walk(match):
            if fn in fset:
                # specified pattern is the exact name
                fset.remove(fn)
            yield fn

        # for dirstate.walk, files=[''] means "walk the whole tree".
        # follow that here, too
        fset.discard(b'')

        for fn in sorted(fset):
            if not self.hasdir(fn):
                match.bad(fn, None)

    def _walk(self, match):
        '''Recursively generates matching file names for walk().'''
        visit = match.visitchildrenset(self._dir[:-1])
        if not visit:
            return

        # yield this dir's files and walk its submanifests
        self._load()
        visit = self._loadchildrensetlazy(visit)
        for p in sorted(list(self._dirs) + list(self._files)):
            if p in self._files:
                fullp = self._subpath(p)
                if match(fullp):
                    yield fullp
            else:
                if not visit or p[:-1] in visit:
                    for f in self._dirs[p]._walk(match):
                        yield f

    def _matches(self, match):
        """recursively generate a new manifest filtered by the match argument."""
        if match.always():
            return self.copy()
        return self._matches_inner(match)

    def _matches_inner(self, match):
        if match.always():
            return self.copy()

        visit = match.visitchildrenset(self._dir[:-1])
        if visit == b'all':
            return self.copy()
        ret = treemanifest(self.nodeconstants, self._dir)
        if not visit:
            return ret

        self._load()
        for fn in self._files:
            # While visitchildrenset *usually* lists only subdirs, this is
            # actually up to the matcher and may have some files in the set().
            # If visit == 'this', we should obviously look at the files in this
            # directory; if visit is a set, and fn is in it, we should inspect
            # fn (but no need to inspect things not in the set).
            if visit != b'this' and fn not in visit:
                continue
            fullp = self._subpath(fn)
            # visitchildrenset isn't perfect, we still need to call the regular
            # matcher code to further filter results.
            if not match(fullp):
                continue
            ret._files[fn] = self._files[fn]
            if fn in self._flags:
                ret._flags[fn] = self._flags[fn]

        visit = self._loadchildrensetlazy(visit)
        for dir, subm in pycompat.iteritems(self._dirs):
            if visit and dir[:-1] not in visit:
                continue
            m = subm._matches_inner(match)
            if not m._isempty():
                ret._dirs[dir] = m

        if not ret._isempty():
            ret._dirty = True
        return ret

    def fastdelta(self, base, changes):
        raise FastdeltaUnavailable()

    def diff(self, m2, match=None, clean=False):
        """Finds changes between the current manifest and m2.

        Args:
          m2: the manifest to which this manifest should be compared.
          clean: if true, include files unchanged between these manifests
                 with a None value in the returned dictionary.

        The result is returned as a dict with filename as key and
        values of the form ((n1,fl1),(n2,fl2)), where n1/n2 is the
        nodeid in the current/other manifest and fl1/fl2 is the flag
        in the current/other manifest. Where the file does not exist,
        the nodeid will be None and the flags will be the empty
        string.
        """
        if match and not match.always():
            m1 = self._matches(match)
            m2 = m2._matches(match)
            return m1.diff(m2, clean=clean)
        result = {}
        emptytree = treemanifest(self.nodeconstants)

        def _iterativediff(t1, t2, stack):
            """compares two tree manifests and append new tree-manifests which
            needs to be compared to stack"""
            if t1._node == t2._node and not t1._dirty and not t2._dirty:
                return
            t1._load()
            t2._load()
            self._loaddifflazy(t1, t2)

            for d, m1 in pycompat.iteritems(t1._dirs):
                m2 = t2._dirs.get(d, emptytree)
                stack.append((m1, m2))

            for d, m2 in pycompat.iteritems(t2._dirs):
                if d not in t1._dirs:
                    stack.append((emptytree, m2))

            for fn, n1 in pycompat.iteritems(t1._files):
                fl1 = t1._flags.get(fn, b'')
                n2 = t2._files.get(fn, None)
                fl2 = t2._flags.get(fn, b'')
                if n1 != n2 or fl1 != fl2:
                    result[t1._subpath(fn)] = ((n1, fl1), (n2, fl2))
                elif clean:
                    result[t1._subpath(fn)] = None

            for fn, n2 in pycompat.iteritems(t2._files):
                if fn not in t1._files:
                    fl2 = t2._flags.get(fn, b'')
                    result[t2._subpath(fn)] = ((None, b''), (n2, fl2))

        stackls = []
        _iterativediff(self, m2, stackls)
        while stackls:
            t1, t2 = stackls.pop()
            # stackls is populated in the function call
            _iterativediff(t1, t2, stackls)
        return result

    def unmodifiedsince(self, m2):
        return not self._dirty and not m2._dirty and self._node == m2._node

    def parse(self, text, readsubtree):
        selflazy = self._lazydirs
        for f, n, fl in _parse(self._nodelen, text):
            if fl == b't':
                f = f + b'/'
                # False below means "doesn't need to be copied" and can use the
                # cached value from readsubtree directly.
                selflazy[f] = (n, readsubtree, False)
            elif b'/' in f:
                # This is a flat manifest, so use __setitem__ and setflag rather
                # than assigning directly to _files and _flags, so we can
                # assign a path in a subdirectory, and to mark dirty (compared
                # to nullid).
                self[f] = n
                if fl:
                    self.setflag(f, fl)
            else:
                # Assigning to _files and _flags avoids marking as dirty,
                # and should be a little faster.
                self._files[f] = n
                if fl:
                    self._flags[f] = fl

    def text(self):
        """Get the full data of this manifest as a bytestring."""
        self._load()
        return _text(self.iterentries())

    def dirtext(self):
        """Get the full data of this directory as a bytestring. Make sure that
        any submanifests have been written first, so their nodeids are correct.
        """
        self._load()
        flags = self.flags
        lazydirs = [
            (d[:-1], v[0], b't') for d, v in pycompat.iteritems(self._lazydirs)
        ]
        dirs = [(d[:-1], self._dirs[d]._node, b't') for d in self._dirs]
        files = [(f, self._files[f], flags(f)) for f in self._files]
        return _text(sorted(dirs + files + lazydirs))

    def read(self, gettext, readsubtree):
        def _load_for_read(s):
            s.parse(gettext(), readsubtree)
            s._dirty = False

        self._loadfunc = _load_for_read

    def writesubtrees(self, m1, m2, writesubtree, match):
        self._load()  # for consistency; should never have any effect here
        m1._load()
        m2._load()
        emptytree = treemanifest(self.nodeconstants)

        def getnode(m, d):
            ld = m._lazydirs.get(d)
            if ld:
                return ld[0]
            return m._dirs.get(d, emptytree)._node

        # let's skip investigating things that `match` says we do not need.
        visit = match.visitchildrenset(self._dir[:-1])
        visit = self._loadchildrensetlazy(visit)
        if visit == b'this' or visit == b'all':
            visit = None
        for d, subm in pycompat.iteritems(self._dirs):
            if visit and d[:-1] not in visit:
                continue
            subp1 = getnode(m1, d)
            subp2 = getnode(m2, d)
            if subp1 == self.nodeconstants.nullid:
                subp1, subp2 = subp2, subp1
            writesubtree(subm, subp1, subp2, match)

    def walksubtrees(self, matcher=None):
        """Returns an iterator of the subtrees of this manifest, including this
        manifest itself.

        If `matcher` is provided, it only returns subtrees that match.
        """
        if matcher and not matcher.visitdir(self._dir[:-1]):
            return
        if not matcher or matcher(self._dir[:-1]):
            yield self

        self._load()
        # OPT: use visitchildrenset to avoid loading everything.
        self._loadalllazy()
        for d, subm in pycompat.iteritems(self._dirs):
            for subtree in subm.walksubtrees(matcher=matcher):
                yield subtree


class manifestfulltextcache(util.lrucachedict):
    """File-backed LRU cache for the manifest cache

    File consists of entries, up to EOF:

    - 20 bytes node, 4 bytes length, <length> manifest data

    These are written in reverse cache order (oldest to newest).

    """

    _file = b'manifestfulltextcache'

    def __init__(self, max):
        super(manifestfulltextcache, self).__init__(max)
        self._dirty = False
        self._read = False
        self._opener = None

    def read(self):
        if self._read or self._opener is None:
            return

        try:
            with self._opener(self._file) as fp:
                set = super(manifestfulltextcache, self).__setitem__
                # ignore trailing data, this is a cache, corruption is skipped
                while True:
                    # TODO do we need to do work here for sha1 portability?
                    node = fp.read(20)
                    if len(node) < 20:
                        break
                    try:
                        size = struct.unpack(b'>L', fp.read(4))[0]
                    except struct.error:
                        break
                    value = bytearray(fp.read(size))
                    if len(value) != size:
                        break
                    set(node, value)
        except IOError:
            # the file is allowed to be missing
            pass

        self._read = True
        self._dirty = False

    def write(self):
        if not self._dirty or self._opener is None:
            return
        # rotate backwards to the first used node
        try:
            with self._opener(
                self._file, b'w', atomictemp=True, checkambig=True
            ) as fp:
                node = self._head.prev
                while True:
                    if node.key in self._cache:
                        fp.write(node.key)
                        fp.write(struct.pack(b'>L', len(node.value)))
                        fp.write(node.value)
                    if node is self._head:
                        break
                    node = node.prev
        except IOError:
            # We could not write the cache (eg: permission error)
            # the content can be missing.
            #
            # We could try harder and see if we could recreate a wcache
            # directory were we coudl write too.
            #
            # XXX the error pass silently, having some way to issue an error
            # log `ui.log` would be nice.
            pass

    def __len__(self):
        if not self._read:
            self.read()
        return super(manifestfulltextcache, self).__len__()

    def __contains__(self, k):
        if not self._read:
            self.read()
        return super(manifestfulltextcache, self).__contains__(k)

    def __iter__(self):
        if not self._read:
            self.read()
        return super(manifestfulltextcache, self).__iter__()

    def __getitem__(self, k):
        if not self._read:
            self.read()
        # the cache lru order can change on read
        setdirty = self._cache.get(k) is not self._head
        value = super(manifestfulltextcache, self).__getitem__(k)
        if setdirty:
            self._dirty = True
        return value

    def __setitem__(self, k, v):
        if not self._read:
            self.read()
        super(manifestfulltextcache, self).__setitem__(k, v)
        self._dirty = True

    def __delitem__(self, k):
        if not self._read:
            self.read()
        super(manifestfulltextcache, self).__delitem__(k)
        self._dirty = True

    def get(self, k, default=None):
        if not self._read:
            self.read()
        return super(manifestfulltextcache, self).get(k, default=default)

    def clear(self, clear_persisted_data=False):
        super(manifestfulltextcache, self).clear()
        if clear_persisted_data:
            self._dirty = True
            self.write()
        self._read = False


# and upper bound of what we expect from compression
# (real live value seems to be "3")
MAXCOMPRESSION = 3


class FastdeltaUnavailable(Exception):
    """Exception raised when fastdelta isn't usable on a manifest."""


@interfaceutil.implementer(repository.imanifeststorage)
class manifestrevlog(object):
    """A revlog that stores manifest texts. This is responsible for caching the
    full-text manifest contents.
    """

    def __init__(
        self,
        nodeconstants,
        opener,
        tree=b'',
        dirlogcache=None,
        treemanifest=False,
    ):
        """Constructs a new manifest revlog

        `indexfile` - used by extensions to have two manifests at once, like
        when transitioning between flatmanifeset and treemanifests.

        `treemanifest` - used to indicate this is a tree manifest revlog. Opener
        options can also be used to make this a tree manifest revlog. The opener
        option takes precedence, so if it is set to True, we ignore whatever
        value is passed in to the constructor.
        """
        self.nodeconstants = nodeconstants
        # During normal operations, we expect to deal with not more than four
        # revs at a time (such as during commit --amend). When rebasing large
        # stacks of commits, the number can go up, hence the config knob below.
        cachesize = 4
        optiontreemanifest = False
        opts = getattr(opener, 'options', None)
        if opts is not None:
            cachesize = opts.get(b'manifestcachesize', cachesize)
            optiontreemanifest = opts.get(b'treemanifest', False)

        self._treeondisk = optiontreemanifest or treemanifest

        self._fulltextcache = manifestfulltextcache(cachesize)

        if tree:
            assert self._treeondisk, b'opts is %r' % opts

        radix = b'00manifest'
        if tree:
            radix = b"meta/" + tree + radix

        self.tree = tree

        # The dirlogcache is kept on the root manifest log
        if tree:
            self._dirlogcache = dirlogcache
        else:
            self._dirlogcache = {b'': self}

        self._revlog = revlog.revlog(
            opener,
            target=(revlog_constants.KIND_MANIFESTLOG, self.tree),
            radix=radix,
            # only root indexfile is cached
            checkambig=not bool(tree),
            mmaplargeindex=True,
            upperboundcomp=MAXCOMPRESSION,
            persistentnodemap=opener.options.get(b'persistent-nodemap', False),
        )

        self.index = self._revlog.index
        self._generaldelta = self._revlog._generaldelta

    def _setupmanifestcachehooks(self, repo):
        """Persist the manifestfulltextcache on lock release"""
        if not util.safehasattr(repo, b'_wlockref'):
            return

        self._fulltextcache._opener = repo.wcachevfs
        if repo._currentlock(repo._wlockref) is None:
            return

        reporef = weakref.ref(repo)
        manifestrevlogref = weakref.ref(self)

        def persistmanifestcache(success):
            # Repo is in an unknown state, do not persist.
            if not success:
                return

            repo = reporef()
            self = manifestrevlogref()
            if repo is None or self is None:
                return
            if repo.manifestlog.getstorage(b'') is not self:
                # there's a different manifest in play now, abort
                return
            self._fulltextcache.write()

        repo._afterlock(persistmanifestcache)

    @property
    def fulltextcache(self):
        return self._fulltextcache

    def clearcaches(self, clear_persisted_data=False):
        self._revlog.clearcaches()
        self._fulltextcache.clear(clear_persisted_data=clear_persisted_data)
        self._dirlogcache = {self.tree: self}

    def dirlog(self, d):
        if d:
            assert self._treeondisk
        if d not in self._dirlogcache:
            mfrevlog = manifestrevlog(
                self.nodeconstants,
                self.opener,
                d,
                self._dirlogcache,
                treemanifest=self._treeondisk,
            )
            self._dirlogcache[d] = mfrevlog
        return self._dirlogcache[d]

    def add(
        self,
        m,
        transaction,
        link,
        p1,
        p2,
        added,
        removed,
        readtree=None,
        match=None,
    ):
        """add some manifest entry in to the manifest log

        input:

          m:           the manifest dict we want to store
          transaction: the open transaction
          p1:          manifest-node of p1
          p2:          manifest-node of p2
          added:       file added/changed compared to parent
          removed:     file removed compared to parent

        tree manifest input:

          readtree:    a function to read a subtree
          match:       a filematcher for the subpart of the tree manifest
        """
        try:
            if p1 not in self.fulltextcache:
                raise FastdeltaUnavailable()
            # If our first parent is in the manifest cache, we can
            # compute a delta here using properties we know about the
            # manifest up-front, which may save time later for the
            # revlog layer.

            _checkforbidden(added)
            # combine the changed lists into one sorted iterator
            work = heapq.merge(
                [(x, False) for x in sorted(added)],
                [(x, True) for x in sorted(removed)],
            )

            arraytext, deltatext = m.fastdelta(self.fulltextcache[p1], work)
            cachedelta = self._revlog.rev(p1), deltatext
            text = util.buffer(arraytext)
            rev = self._revlog.addrevision(
                text, transaction, link, p1, p2, cachedelta
            )
            n = self._revlog.node(rev)
        except FastdeltaUnavailable:
            # The first parent manifest isn't already loaded or the
            # manifest implementation doesn't support fastdelta, so
            # we'll just encode a fulltext of the manifest and pass
            # that through to the revlog layer, and let it handle the
            # delta process.
            if self._treeondisk:
                assert readtree, b"readtree must be set for treemanifest writes"
                assert match, b"match must be specified for treemanifest writes"
                m1 = readtree(self.tree, p1)
                m2 = readtree(self.tree, p2)
                n = self._addtree(
                    m, transaction, link, m1, m2, readtree, match=match
                )
                arraytext = None
            else:
                text = m.text()
                rev = self._revlog.addrevision(text, transaction, link, p1, p2)
                n = self._revlog.node(rev)
                arraytext = bytearray(text)

        if arraytext is not None:
            self.fulltextcache[n] = arraytext

        return n

    def _addtree(self, m, transaction, link, m1, m2, readtree, match):
        # If the manifest is unchanged compared to one parent,
        # don't write a new revision
        if self.tree != b'' and (
            m.unmodifiedsince(m1) or m.unmodifiedsince(m2)
        ):
            return m.node()

        def writesubtree(subm, subp1, subp2, match):
            sublog = self.dirlog(subm.dir())
            sublog.add(
                subm,
                transaction,
                link,
                subp1,
                subp2,
                None,
                None,
                readtree=readtree,
                match=match,
            )

        m.writesubtrees(m1, m2, writesubtree, match)
        text = m.dirtext()
        n = None
        if self.tree != b'':
            # Double-check whether contents are unchanged to one parent
            if text == m1.dirtext():
                n = m1.node()
            elif text == m2.dirtext():
                n = m2.node()

        if not n:
            rev = self._revlog.addrevision(
                text, transaction, link, m1.node(), m2.node()
            )
            n = self._revlog.node(rev)

        # Save nodeid so parent manifest can calculate its nodeid
        m.setnode(n)
        return n

    def __len__(self):
        return len(self._revlog)

    def __iter__(self):
        return self._revlog.__iter__()

    def rev(self, node):
        return self._revlog.rev(node)

    def node(self, rev):
        return self._revlog.node(rev)

    def lookup(self, value):
        return self._revlog.lookup(value)

    def parentrevs(self, rev):
        return self._revlog.parentrevs(rev)

    def parents(self, node):
        return self._revlog.parents(node)

    def linkrev(self, rev):
        return self._revlog.linkrev(rev)

    def checksize(self):
        return self._revlog.checksize()

    def revision(self, node, _df=None, raw=False):
        return self._revlog.revision(node, _df=_df, raw=raw)

    def rawdata(self, node, _df=None):
        return self._revlog.rawdata(node, _df=_df)

    def revdiff(self, rev1, rev2):
        return self._revlog.revdiff(rev1, rev2)

    def cmp(self, node, text):
        return self._revlog.cmp(node, text)

    def deltaparent(self, rev):
        return self._revlog.deltaparent(rev)

    def emitrevisions(
        self,
        nodes,
        nodesorder=None,
        revisiondata=False,
        assumehaveparentrevisions=False,
        deltamode=repository.CG_DELTAMODE_STD,
        sidedata_helpers=None,
    ):
        return self._revlog.emitrevisions(
            nodes,
            nodesorder=nodesorder,
            revisiondata=revisiondata,
            assumehaveparentrevisions=assumehaveparentrevisions,
            deltamode=deltamode,
            sidedata_helpers=sidedata_helpers,
        )

    def addgroup(
        self,
        deltas,
        linkmapper,
        transaction,
        alwayscache=False,
        addrevisioncb=None,
        duplicaterevisioncb=None,
    ):
        return self._revlog.addgroup(
            deltas,
            linkmapper,
            transaction,
            alwayscache=alwayscache,
            addrevisioncb=addrevisioncb,
            duplicaterevisioncb=duplicaterevisioncb,
        )

    def rawsize(self, rev):
        return self._revlog.rawsize(rev)

    def getstrippoint(self, minlink):
        return self._revlog.getstrippoint(minlink)

    def strip(self, minlink, transaction):
        return self._revlog.strip(minlink, transaction)

    def files(self):
        return self._revlog.files()

    def clone(self, tr, destrevlog, **kwargs):
        if not isinstance(destrevlog, manifestrevlog):
            raise error.ProgrammingError(b'expected manifestrevlog to clone()')

        return self._revlog.clone(tr, destrevlog._revlog, **kwargs)

    def storageinfo(
        self,
        exclusivefiles=False,
        sharedfiles=False,
        revisionscount=False,
        trackedsize=False,
        storedsize=False,
    ):
        return self._revlog.storageinfo(
            exclusivefiles=exclusivefiles,
            sharedfiles=sharedfiles,
            revisionscount=revisionscount,
            trackedsize=trackedsize,
            storedsize=storedsize,
        )

    @property
    def opener(self):
        return self._revlog.opener

    @opener.setter
    def opener(self, value):
        self._revlog.opener = value


@interfaceutil.implementer(repository.imanifestlog)
class manifestlog(object):
    """A collection class representing the collection of manifest snapshots
    referenced by commits in the repository.

    In this situation, 'manifest' refers to the abstract concept of a snapshot
    of the list of files in the given commit. Consumers of the output of this
    class do not care about the implementation details of the actual manifests
    they receive (i.e. tree or flat or lazily loaded, etc)."""

    def __init__(self, opener, repo, rootstore, narrowmatch):
        self.nodeconstants = repo.nodeconstants
        usetreemanifest = False
        cachesize = 4

        opts = getattr(opener, 'options', None)
        if opts is not None:
            usetreemanifest = opts.get(b'treemanifest', usetreemanifest)
            cachesize = opts.get(b'manifestcachesize', cachesize)

        self._treemanifests = usetreemanifest

        self._rootstore = rootstore
        self._rootstore._setupmanifestcachehooks(repo)
        self._narrowmatch = narrowmatch

        # A cache of the manifestctx or treemanifestctx for each directory
        self._dirmancache = {}
        self._dirmancache[b''] = util.lrucachedict(cachesize)

        self._cachesize = cachesize

    def __getitem__(self, node):
        """Retrieves the manifest instance for the given node. Throws a
        LookupError if not found.
        """
        return self.get(b'', node)

    def get(self, tree, node, verify=True):
        """Retrieves the manifest instance for the given node. Throws a
        LookupError if not found.

        `verify` - if True an exception will be thrown if the node is not in
                   the revlog
        """
        if node in self._dirmancache.get(tree, ()):
            return self._dirmancache[tree][node]

        if not self._narrowmatch.always():
            if not self._narrowmatch.visitdir(tree[:-1]):
                return excludeddirmanifestctx(self.nodeconstants, tree, node)
        if tree:
            if self._rootstore._treeondisk:
                if verify:
                    # Side-effect is LookupError is raised if node doesn't
                    # exist.
                    self.getstorage(tree).rev(node)

                m = treemanifestctx(self, tree, node)
            else:
                raise error.Abort(
                    _(
                        b"cannot ask for manifest directory '%s' in a flat "
                        b"manifest"
                    )
                    % tree
                )
        else:
            if verify:
                # Side-effect is LookupError is raised if node doesn't exist.
                self._rootstore.rev(node)

            if self._treemanifests:
                m = treemanifestctx(self, b'', node)
            else:
                m = manifestctx(self, node)

        if node != self.nodeconstants.nullid:
            mancache = self._dirmancache.get(tree)
            if not mancache:
                mancache = util.lrucachedict(self._cachesize)
                self._dirmancache[tree] = mancache
            mancache[node] = m
        return m

    def getstorage(self, tree):
        return self._rootstore.dirlog(tree)

    def clearcaches(self, clear_persisted_data=False):
        self._dirmancache.clear()
        self._rootstore.clearcaches(clear_persisted_data=clear_persisted_data)

    def rev(self, node):
        return self._rootstore.rev(node)

    def update_caches(self, transaction):
        return self._rootstore._revlog.update_caches(transaction=transaction)


@interfaceutil.implementer(repository.imanifestrevisionwritable)
class memmanifestctx(object):
    def __init__(self, manifestlog):
        self._manifestlog = manifestlog
        self._manifestdict = manifestdict(manifestlog.nodeconstants.nodelen)

    def _storage(self):
        return self._manifestlog.getstorage(b'')

    def copy(self):
        memmf = memmanifestctx(self._manifestlog)
        memmf._manifestdict = self.read().copy()
        return memmf

    def read(self):
        return self._manifestdict

    def write(self, transaction, link, p1, p2, added, removed, match=None):
        return self._storage().add(
            self._manifestdict,
            transaction,
            link,
            p1,
            p2,
            added,
            removed,
            match=match,
        )


@interfaceutil.implementer(repository.imanifestrevisionstored)
class manifestctx(object):
    """A class representing a single revision of a manifest, including its
    contents, its parent revs, and its linkrev.
    """

    def __init__(self, manifestlog, node):
        self._manifestlog = manifestlog
        self._data = None

        self._node = node

        # TODO: We eventually want p1, p2, and linkrev exposed on this class,
        # but let's add it later when something needs it and we can load it
        # lazily.
        # self.p1, self.p2 = store.parents(node)
        # rev = store.rev(node)
        # self.linkrev = store.linkrev(rev)

    def _storage(self):
        return self._manifestlog.getstorage(b'')

    def node(self):
        return self._node

    def copy(self):
        memmf = memmanifestctx(self._manifestlog)
        memmf._manifestdict = self.read().copy()
        return memmf

    @propertycache
    def parents(self):
        return self._storage().parents(self._node)

    def read(self):
        if self._data is None:
            nc = self._manifestlog.nodeconstants
            if self._node == nc.nullid:
                self._data = manifestdict(nc.nodelen)
            else:
                store = self._storage()
                if self._node in store.fulltextcache:
                    text = pycompat.bytestr(store.fulltextcache[self._node])
                else:
                    text = store.revision(self._node)
                    arraytext = bytearray(text)
                    store.fulltextcache[self._node] = arraytext
                self._data = manifestdict(nc.nodelen, text)
        return self._data

    def readfast(self, shallow=False):
        """Calls either readdelta or read, based on which would be less work.
        readdelta is called if the delta is against the p1, and therefore can be
        read quickly.

        If `shallow` is True, nothing changes since this is a flat manifest.
        """
        store = self._storage()
        r = store.rev(self._node)
        deltaparent = store.deltaparent(r)
        if deltaparent != nullrev and deltaparent in store.parentrevs(r):
            return self.readdelta()
        return self.read()

    def readdelta(self, shallow=False):
        """Returns a manifest containing just the entries that are present
        in this manifest, but not in its p1 manifest. This is efficient to read
        if the revlog delta is already p1.

        Changing the value of `shallow` has no effect on flat manifests.
        """
        store = self._storage()
        r = store.rev(self._node)
        d = mdiff.patchtext(store.revdiff(store.deltaparent(r), r))
        return manifestdict(store.nodeconstants.nodelen, d)

    def find(self, key):
        return self.read().find(key)


@interfaceutil.implementer(repository.imanifestrevisionwritable)
class memtreemanifestctx(object):
    def __init__(self, manifestlog, dir=b''):
        self._manifestlog = manifestlog
        self._dir = dir
        self._treemanifest = treemanifest(manifestlog.nodeconstants)

    def _storage(self):
        return self._manifestlog.getstorage(b'')

    def copy(self):
        memmf = memtreemanifestctx(self._manifestlog, dir=self._dir)
        memmf._treemanifest = self._treemanifest.copy()
        return memmf

    def read(self):
        return self._treemanifest

    def write(self, transaction, link, p1, p2, added, removed, match=None):
        def readtree(dir, node):
            return self._manifestlog.get(dir, node).read()

        return self._storage().add(
            self._treemanifest,
            transaction,
            link,
            p1,
            p2,
            added,
            removed,
            readtree=readtree,
            match=match,
        )


@interfaceutil.implementer(repository.imanifestrevisionstored)
class treemanifestctx(object):
    def __init__(self, manifestlog, dir, node):
        self._manifestlog = manifestlog
        self._dir = dir
        self._data = None

        self._node = node

        # TODO: Load p1/p2/linkrev lazily. They need to be lazily loaded so that
        # we can instantiate treemanifestctx objects for directories we don't
        # have on disk.
        # self.p1, self.p2 = store.parents(node)
        # rev = store.rev(node)
        # self.linkrev = store.linkrev(rev)

    def _storage(self):
        narrowmatch = self._manifestlog._narrowmatch
        if not narrowmatch.always():
            if not narrowmatch.visitdir(self._dir[:-1]):
                return excludedmanifestrevlog(
                    self._manifestlog.nodeconstants, self._dir
                )
        return self._manifestlog.getstorage(self._dir)

    def read(self):
        if self._data is None:
            store = self._storage()
            if self._node == self._manifestlog.nodeconstants.nullid:
                self._data = treemanifest(self._manifestlog.nodeconstants)
            # TODO accessing non-public API
            elif store._treeondisk:
                m = treemanifest(self._manifestlog.nodeconstants, dir=self._dir)

                def gettext():
                    return store.revision(self._node)

                def readsubtree(dir, subm):
                    # Set verify to False since we need to be able to create
                    # subtrees for trees that don't exist on disk.
                    return self._manifestlog.get(dir, subm, verify=False).read()

                m.read(gettext, readsubtree)
                m.setnode(self._node)
                self._data = m
            else:
                if self._node in store.fulltextcache:
                    text = pycompat.bytestr(store.fulltextcache[self._node])
                else:
                    text = store.revision(self._node)
                    arraytext = bytearray(text)
                    store.fulltextcache[self._node] = arraytext
                self._data = treemanifest(
                    self._manifestlog.nodeconstants, dir=self._dir, text=text
                )

        return self._data

    def node(self):
        return self._node

    def copy(self):
        memmf = memtreemanifestctx(self._manifestlog, dir=self._dir)
        memmf._treemanifest = self.read().copy()
        return memmf

    @propertycache
    def parents(self):
        return self._storage().parents(self._node)

    def readdelta(self, shallow=False):
        """Returns a manifest containing just the entries that are present
        in this manifest, but not in its p1 manifest. This is efficient to read
        if the revlog delta is already p1.

        If `shallow` is True, this will read the delta for this directory,
        without recursively reading subdirectory manifests. Instead, any
        subdirectory entry will be reported as it appears in the manifest, i.e.
        the subdirectory will be reported among files and distinguished only by
        its 't' flag.
        """
        store = self._storage()
        if shallow:
            r = store.rev(self._node)
            d = mdiff.patchtext(store.revdiff(store.deltaparent(r), r))
            return manifestdict(store.nodeconstants.nodelen, d)
        else:
            # Need to perform a slow delta
            r0 = store.deltaparent(store.rev(self._node))
            m0 = self._manifestlog.get(self._dir, store.node(r0)).read()
            m1 = self.read()
            md = treemanifest(self._manifestlog.nodeconstants, dir=self._dir)
            for f, ((n0, fl0), (n1, fl1)) in pycompat.iteritems(m0.diff(m1)):
                if n1:
                    md[f] = n1
                    if fl1:
                        md.setflag(f, fl1)
            return md

    def readfast(self, shallow=False):
        """Calls either readdelta or read, based on which would be less work.
        readdelta is called if the delta is against the p1, and therefore can be
        read quickly.

        If `shallow` is True, it only returns the entries from this manifest,
        and not any submanifests.
        """
        store = self._storage()
        r = store.rev(self._node)
        deltaparent = store.deltaparent(r)
        if deltaparent != nullrev and deltaparent in store.parentrevs(r):
            return self.readdelta(shallow=shallow)

        if shallow:
            return manifestdict(
                store.nodeconstants.nodelen, store.revision(self._node)
            )
        else:
            return self.read()

    def find(self, key):
        return self.read().find(key)


class excludeddir(treemanifest):
    """Stand-in for a directory that is excluded from the repository.

    With narrowing active on a repository that uses treemanifests,
    some of the directory revlogs will be excluded from the resulting
    clone. This is a huge storage win for clients, but means we need
    some sort of pseudo-manifest to surface to internals so we can
    detect a merge conflict outside the narrowspec. That's what this
    class is: it stands in for a directory whose node is known, but
    whose contents are unknown.
    """

    def __init__(self, nodeconstants, dir, node):
        super(excludeddir, self).__init__(nodeconstants, dir)
        self._node = node
        # Add an empty file, which will be included by iterators and such,
        # appearing as the directory itself (i.e. something like "dir/")
        self._files[b''] = node
        self._flags[b''] = b't'

    # Manifests outside the narrowspec should never be modified, so avoid
    # copying. This makes a noticeable difference when there are very many
    # directories outside the narrowspec. Also, it makes sense for the copy to
    # be of the same type as the original, which would not happen with the
    # super type's copy().
    def copy(self):
        return self


class excludeddirmanifestctx(treemanifestctx):
    """context wrapper for excludeddir - see that docstring for rationale"""

    def __init__(self, nodeconstants, dir, node):
        self.nodeconstants = nodeconstants
        self._dir = dir
        self._node = node

    def read(self):
        return excludeddir(self.nodeconstants, self._dir, self._node)

    def readfast(self, shallow=False):
        # special version of readfast since we don't have underlying storage
        return self.read()

    def write(self, *args):
        raise error.ProgrammingError(
            b'attempt to write manifest from excluded dir %s' % self._dir
        )


class excludedmanifestrevlog(manifestrevlog):
    """Stand-in for excluded treemanifest revlogs.

    When narrowing is active on a treemanifest repository, we'll have
    references to directories we can't see due to the revlog being
    skipped. This class exists to conform to the manifestrevlog
    interface for those directories and proactively prevent writes to
    outside the narrowspec.
    """

    def __init__(self, nodeconstants, dir):
        self.nodeconstants = nodeconstants
        self._dir = dir

    def __len__(self):
        raise error.ProgrammingError(
            b'attempt to get length of excluded dir %s' % self._dir
        )

    def rev(self, node):
        raise error.ProgrammingError(
            b'attempt to get rev from excluded dir %s' % self._dir
        )

    def linkrev(self, node):
        raise error.ProgrammingError(
            b'attempt to get linkrev from excluded dir %s' % self._dir
        )

    def node(self, rev):
        raise error.ProgrammingError(
            b'attempt to get node from excluded dir %s' % self._dir
        )

    def add(self, *args, **kwargs):
        # We should never write entries in dirlogs outside the narrow clone.
        # However, the method still gets called from writesubtree() in
        # _addtree(), so we need to handle it. We should possibly make that
        # avoid calling add() with a clean manifest (_dirty is always False
        # in excludeddir instances).
        pass
